"""
This module provides functionality to retrieve detailed information about orders from the Shopify store.
It supports filtering orders by customer ID, order IDs, and order names.

Module Name: get_order_details

This file contains the code for retrieving detailed order information from Shopify.
"""

import inspect
import json

import shopify
from pydantic import BaseModel

from arklex.env.tools.shopify._exception_prompt import ShopifyExceptionPrompt
from arklex.env.tools.shopify.base.entities import ShopifyAdminAuth
from arklex.env.tools.shopify.utils.utils import authorify_admin
from arklex.env.tools.shopify.utils.utils_slots import (
    ShopifyGetOrderDetailsSlots,
)
from arklex.env.tools.tools import register_tool
from arklex.utils.exceptions import ToolExecutionError


class GetOrderDetailsOutput(BaseModel):
    message_flow: str


@register_tool(
    description="Get the status and details of an order.",
    slots=ShopifyGetOrderDetailsSlots.get_all_slots(),
)
def get_order_details(
    order_ids: list[str],
    order_names: list[str],
    user_id: str,
    auth: ShopifyAdminAuth,
    limit: int = 10,
    **kwargs: object,
) -> GetOrderDetailsOutput:
    """
    Retrieve detailed information about orders from the Shopify store.

    Args:
        order_ids (List[str]): List of order IDs to filter by.
        order_names (List[str]): List of order names to filter by.
        user_id (str): The customer ID to filter orders by.
        limit (int, optional): Maximum number of orders to return. Defaults to 10.
        **kwargs (GetOrderDetailsParams): Additional keyword arguments for authentication.

    Returns:
        GetOrderDetailsOutput: A formatted string containing detailed information about each order, including:
            - Order ID and name
            - Creation and cancellation dates
            - Return status
            - Status page URL
            - Total price
            - Fulfillment status
            - Line items with their details
        str: "You have no orders placed." if no orders are found.

    Raises:
        ToolExecutionError: If there's an error retrieving the orders.
    """
    func_name = inspect.currentframe().f_code.co_name
    auth = authorify_admin(auth)
    limit = int(limit) if limit else 10

    try:
        query = f"customer_id:{user_id.split('/')[-1]}"
        if order_ids:
            order_ids = " OR ".join(f"id:{oid.split('/')[-1]}" for oid in order_ids)
            query += f" AND ({order_ids})"
        if order_names:
            order_names = " OR ".join(f"name:{name}" for name in order_names)
            query += f" AND ({order_names})"
        with shopify.Session.temp(**auth):
            response = shopify.GraphQL().execute(f"""
            {{
                orders (first: {limit}, query:"{query}") {{
                    nodes {{
                        id
                        name
                        createdAt
                        cancelledAt
                        returnStatus
                        statusPageUrl
                        totalPriceSet {{
                            presentmentMoney {{
                                amount
                            }}
                        }}
                        fulfillments {{
                            displayStatus
                            trackingInfo {{
                                number
                                url
                            }}
                        }}
                        lineItems(first: 10) {{
                            edges {{
                                node {{
                                    id
                                    title
                                    quantity
                                    variant {{
                                        id
                                        product {{
                                            id
                                        }}
                                    }}
                                }}
                            }}
                        }}
                    }}
                }}
            }}
            """)
            result = json.loads(response)["data"]["orders"]["nodes"]
            if len(result) == 0:
                return GetOrderDetailsOutput(
                    message_flow="You have no orders placed.",
                )
            response_text = ""
            for order in result:
                response_text += f"Order ID: {order.get('id', 'None')}\n"
                response_text += f"Order Name: {order.get('name', 'None')}\n"
                response_text += f"Created At: {order.get('createdAt', 'None')}\n"
                response_text += f"Cancelled At: {order.get('cancelledAt', 'None')}\n"
                response_text += f"Return Status: {order.get('returnStatus', 'None')}\n"
                response_text += (
                    f"Status Page URL: {order.get('statusPageUrl', 'None')}\n"
                )
                response_text += f"Total Price: {order.get('totalPriceSet', {}).get('presentmentMoney', {}).get('amount', 'None')}\n"
                response_text += (
                    f"Fulfillment Status: {order.get('fulfillments', 'None')}\n"
                )
                response_text += "Line Items:\n"
                for item in order.get("lineItems", {}).get("edges", []):
                    response_text += (
                        f"    Title: {item.get('node', {}).get('title', 'None')}\n"
                    )
                    response_text += f"    Quantity: {item.get('node', {}).get('quantity', 'None')}\n"
                    response_text += (
                        f"    Variant: {item.get('node', {}).get('variant', {})}\n"
                    )
                response_text += "\n"
        return GetOrderDetailsOutput(
            message_flow=response_text,
        )
    except Exception as e:
        raise ToolExecutionError(
            func_name,
            extra_message=ShopifyExceptionPrompt.ORDERS_NOT_FOUND_PROMPT,
        ) from e



"""Entities for task graph management.

This module defines the entities used in the task graph management system,
including PathNode, NodeTypeEnum, NodeInfo, and Taskgraph.

Key Components:
- PathNode: Represents a node in the processing path of the conversation graph.
- NodeTypeEnum: Represents the type of node.
- NodeInfo: Represents the information about a node.
- Taskgraph: Represents the overall structure of the task processing graph.
"""

from enum import Enum
from typing import Any

from pydantic import BaseModel, Field

from arklex.orchestrator.entities.orchestrator_state_entities import StatusEnum
from arklex.orchestrator.NLU.entities.slot_entities import Slot


class NodeTypeEnum(str, Enum):
    """Enumeration of node types.

    This enum defines the different types of nodes that can exist in the
    conversation graph, each with specific processing requirements.

    The enum provides:
    1. Clear node type definitions
    2. Type-safe node type handling
    3. Consistent node type tracking

    Values:
        NONE: No specific type.
        START: Starting node.
        MULTIPLE_CHOICE: Node with multiple choice options.
    """

    NONE: str = ""
    START: str = "start"
    MULTIPLE_CHOICE: str = "multiple_choice"


class NodeInfo(BaseModel):
    """Information about a processing node.

    This class contains detailed information about a node in the processing graph,
    including its type, resource information, and processing attributes.

    The class provides:
    1. Node identification
    2. Type management
    3. Resource tracking
    4. Processing attributes
    5. Type-safe node information management

    Attributes:
        node_id (Optional[str]): Unique identifier for the node.
        resource (Dict[str, str]): Resource information.
        attribute (Dict[str, Any]): Node attributes.
        data (Dict[str, Any]): Node data.
        is_leaf (bool): Whether the node is a leaf node.
        add_flow_stack (Optional[bool]): Whether to add to flow stack.
        additional_args (Optional[Dict[str, Any]]): Additional arguments for the node.
    """

    node_id: str = Field(default="")
    resource: dict[str, str] = Field(default_factory=dict)
    attribute: dict[str, Any] = Field(default_factory=dict)
    data: dict[str, Any] = Field(default_factory=dict)
    successors: list = Field(default_factory=list)
    predecessors: list = Field(default_factory=list)
    is_leaf: bool = Field(default=False)
    add_flow_stack: bool | None = Field(default=False)


class PathNode(BaseModel):
    """Node in the processing path.

    This class represents a node in the processing path of the conversation graph,
    tracking its state and relationships with other nodes.

    The class provides:
    1. Node identification
    2. State tracking
    3. Flow stack management
    4. Nested graph handling
    5. Intent tracking
    6. Type-safe node management

    Attributes:
        node_id (str): Unique identifier for the node.
        is_skipped (bool): Whether the node was skipped.
        in_flow_stack (bool): Whether the node is in the flow stack.
        nested_graph_node_value (Optional[str]): Value for nested graph nodes.
        nested_graph_leaf_jump (Optional[int]): Jump value for nested graph leaves.
        global_intent (str): Global intent associated with the node.
    """

    node_id: str
    is_skipped: bool = False
    in_flow_stack: bool = False
    nested_graph_node_value: str | None = None
    nested_graph_leaf_jump: int | None = None
    global_intent: str = Field(default="")


class Taskgraph(BaseModel):
    """Graph structure for task processing.

    This class manages the overall structure of the task processing graph,
    including node states, paths, and dialogue states.

    The class provides:
    1. Dialogue state management
    2. Path tracking
    3. Node status management
    4. Intent handling
    5. Node limit enforcement
    6. NLU record tracking
    7. Type-safe graph management

    Attributes:
        dialog_states (Dict[str, List[Slot]]): States of dialogue slots.
        path (List[PathNode]): Processing path nodes.
        curr_node (str): Current node identifier.
        intent (str): Current intent.
        curr_global_intent (str): Current global intent.
        node_limit (Dict[str, int]): Limits for node processing.
        nlu_records (List[Any]): Natural language understanding records.
        node_status (Dict[str, StatusEnum]): Status of each node.
        available_global_intents (List[str]): List of available global intents.
    """

    # Need add global intent
    dialog_states: dict[str, list[Slot]] = Field(default_factory=dict)
    path: list[PathNode] = Field(default_factory=list)
    curr_node: str = Field(default="")
    intent: str = Field(default="")
    curr_global_intent: str = Field(default="")
    node_limit: dict[str, int] = Field(default_factory=dict)
    nlu_records: list[Any] = Field(default_factory=list)
    node_status: dict[str, StatusEnum] = Field(default_factory=dict)
    available_global_intents: list[str] = Field(default_factory=list)



"""Evaluation package for the Arklex framework.

This package provides tools and utilities for evaluating the performance and effectiveness
of the Arklex framework. It includes functionality for conversation simulation, user profile
generation, document processing, and metrics extraction. The package supports both first-pass
and second-pass conversation evaluation, with capabilities for synthetic data generation,
goal completion tracking, and performance analysis.

Key Components:
- Conversation Simulation:
  - First-pass conversation simulation (simulate_first_pass_convos.py)
  - Second-pass conversation simulation (simulate_second_pass_convos.py)
  - Conversation information extraction (extract_conversation_info.py)
- User Profile Management:
  - User profile generation (build_user_profiles.py)
  - User attributes configuration (user_attributes.json)
- Document Processing:
  - Document retrieval and processing (get_documents.py)
  - ChatGPT utilities for evaluation (chatgpt_utils.py)
- Data Management:
  - Data directory for storing evaluation data and results
  - README with evaluation guidelines and procedures

Key Features:
- Comprehensive conversation simulation
- Synthetic user profile generation
- Document processing and analysis
- Performance metrics extraction
- Goal completion tracking
- Automated evaluation workflows

Usage:
    # Import evaluation utilities as needed
    from arklex.evaluation import (
        simulate_first_pass_convos,
        simulate_second_pass_convos,
        extract_conversation_info,
        build_user_profiles,
        get_documents,
        chatgpt_utils
    )

    # Run evaluation workflows
    # 1. Generate user profiles
    # 2. Simulate conversations
    # 3. Extract and analyze metrics
    # 4. Generate evaluation reports
"""

# Import modules to make them available at package level
from . import extract_conversation_info as extract_conversation_info

# Note: This package contains tools for comprehensive evaluation of the Arklex framework,
# including conversation simulation, user profile generation, and performance analysis.



"""Benchmark runner for the TAU benchmark.

This module provides functionality for running and evaluating agents on the TAU benchmark,
including support for different environments (retail, airline), user strategies, and
model providers. It handles task execution, result collection, and metric calculation
for evaluating agent performance.

The module includes:
- Task execution and result collection
- Performance metric calculation (average reward, pass^k)
- Concurrent task execution with configurable concurrency
- Checkpointing and result persistence
- Environment and agent management
"""

import json
import multiprocessing
import os
import random
import traceback
from concurrent.futures import ThreadPoolExecutor
from math import comb
from typing import Any

from litellm import provider_list

from benchmark.tau_bench.agents.base import Agent
from benchmark.tau_bench.envs import get_env
from benchmark.tau_bench.envs.user import UserStrategy
from benchmark.tau_bench.tau_types import EnvRunResult, RunConfig


def run(config: RunConfig) -> list[EnvRunResult]:
    """Run the TAU benchmark with the specified configuration.

    This function executes the benchmark tasks according to the provided configuration,
    including environment setup, task execution, and result collection. It supports
    multiple trials, task shuffling, and concurrent execution. Results are saved to
    a checkpoint file for persistence and analysis.

    Args:
        config (RunConfig): Configuration for the benchmark run, including environment,
            user strategy, model provider, and task settings.

    Returns:
        List[EnvRunResult]: List of results from all task executions.

    Raises:
        AssertionError: If the environment, user model provider, task split, or
            user strategy is invalid.
    """
    assert config.env in ["retail", "airline"], (
        "Only retail and airline envs are supported"
    )
    assert config.user_model_provider in provider_list, "Invalid user model provider"
    assert config.task_split in ["train", "test", "dev"], "Invalid task split"
    assert config.user_strategy in [item.value for item in UserStrategy], (
        "Invalid user strategy"
    )

    random.seed(config.seed)
    ckpt_path: str = f"{config.output_dir}/tau_bench_evaluation.json"
    if not os.path.exists(config.log_dir):
        os.makedirs(config.log_dir)

    print(f"Loading user with strategy: {config.user_strategy}")
    env: Any = get_env(
        config.env,
        user_strategy=config.user_strategy,
        user_model=config.user_model,
        user_provider=config.user_model_provider,
        task_split=config.task_split,
    )
    agent: Agent = agent_factory(config=config)
    end_index: int = (
        len(env.tasks)
        if config.end_index == -1
        else min(config.end_index, len(env.tasks))
    )
    results: list[EnvRunResult] = []
    lock: multiprocessing.Lock = multiprocessing.Lock()
    if config.task_ids and len(config.task_ids) > 0:
        print(f"Running tasks {config.task_ids} (checkpoint path: {ckpt_path})")
    else:
        print(
            f"Running tasks {config.start_index} to {end_index} (checkpoint path: {ckpt_path})"
        )
    for i in range(config.num_trials):
        if config.task_ids and len(config.task_ids) > 0:
            idxs: list[int] = config.task_ids
        else:
            idxs: list[int] = list(range(config.start_index, end_index))
        if config.shuffle:
            random.shuffle(idxs)

        def _run(idx: int, trial_num: int = i) -> EnvRunResult:
            isolated_env: Any = get_env(
                config.env,
                user_strategy=config.user_strategy,
                user_model=config.user_model,
                task_split=config.task_split,
                user_provider=config.user_model_provider,
                task_index=idx,
            )

            print(f"Running task {idx}")
            try:
                res: Any = agent.solve(
                    env=isolated_env,
                    task_index=idx,
                )
                result: EnvRunResult = EnvRunResult(
                    task_id=idx,
                    reward=res.reward,
                    info=res.info,
                    traj=res.messages,
                    trial=trial_num,
                )
            except Exception as e:
                result: EnvRunResult = EnvRunResult(
                    task_id=idx,
                    reward=0.0,
                    info={"error": str(e), "traceback": traceback.format_exc()},
                    traj=[],
                    trial=trial_num,
                )
            print(
                "✅" if result.reward == 1 else "❌",
                f"task_id={idx}",
                result.info,
            )
            print("-----")
            with lock:
                data: list[Any] = []
                with open(ckpt_path, "w") as f:
                    json.dump(data + [result.model_dump()], f, indent=2)
            return result

        with ThreadPoolExecutor(max_workers=config.max_concurrency) as executor:
            res: list[EnvRunResult] = list(executor.map(_run, idxs))
            results.extend(res)

    avg_reward: float
    pass_hat_ks: dict[int, float]
    avg_reward, pass_hat_ks = get_metrics(results)
    display_metrics(avg_reward, pass_hat_ks)

    tasks_res: list[dict[str, Any]] = [result.model_dump() for result in results]
    with open(ckpt_path, "w") as f:
        tau_bench_evaluation: dict[str, Any] = {
            "average_reward": avg_reward,
            "pass_k": pass_hat_ks,
            "task_results": tasks_res,
        }
        json.dump(tau_bench_evaluation, f, indent=2)
        print(f"\n📄 Results saved to {ckpt_path}\n")
    return results


def agent_factory(config: RunConfig) -> Agent:
    """Create an agent instance based on the configuration.

    This function creates and returns an appropriate agent instance for the benchmark
    based on the provided configuration. Currently, it creates an AgentFirstOrg
    instance with the specified taskgraph directory.

    Args:
        config (RunConfig): Configuration for the agent, including taskgraph directory.

    Returns:
        Agent: An instance of the appropriate agent class.
    """
    from benchmark.tau_bench.agents.agent_first_org import AgentFirstOrg

    return AgentFirstOrg(taskgraph_dir=config.taskgraph_dir)


def get_metrics(results: list[EnvRunResult]) -> tuple[float, dict[int, float]]:
    """Calculate performance metrics from benchmark results.

    This function calculates the average reward and pass^k metrics from the benchmark
    results. The pass^k metric represents the probability of passing k trials, which
    is calculated using the binomial coefficient formula from the TAU benchmark paper.

    Args:
        results (List[EnvRunResult]): List of results from task executions.

    Returns:
        Tuple[float, Dict[int, float]]: A tuple containing:
            - float: The average reward across all tasks
            - Dict[int, float]: A dictionary mapping k to pass^k values
    """

    def is_successful(reward: float) -> bool:
        return (1 - 1e-6) <= reward <= (1 + 1e-6)

    num_trials: int = len({r.trial for r in results})
    rewards: list[float] = [r.reward for r in results]
    avg_reward: float = sum(rewards) / len(rewards)
    # c from https://arxiv.org/pdf/2406.12045
    c_per_task_id: dict[int, int] = {}
    for result in results:
        if result.task_id not in c_per_task_id:
            c_per_task_id[result.task_id] = 1 if is_successful(result.reward) else 0
        else:
            c_per_task_id[result.task_id] += 1 if is_successful(result.reward) else 0
    pass_hat_ks: dict[int, float] = {}
    for k in range(1, num_trials + 1):
        sum_task_pass_hat_k: float = 0
        for c in c_per_task_id.values():
            sum_task_pass_hat_k += comb(c, k) / comb(num_trials, k)
        pass_hat_ks[k] = sum_task_pass_hat_k / len(c_per_task_id)
    return avg_reward, pass_hat_ks


def display_metrics(avg_reward: float, pass_hat_ks: dict[int, float]) -> None:
    """Display benchmark performance metrics.

    This function prints the average reward and pass^k metrics in a formatted way,
    providing a clear overview of the benchmark performance.

    Args:
        avg_reward (float): The average reward across all tasks.
        pass_hat_ks (Dict[int, float]): Dictionary mapping k to pass^k values.
    """
    print(f"🏆 Average reward: {avg_reward}")
    print("📈 Pass^k")
    for k, pass_hat_k in pass_hat_ks.items():
        print(f"  k={k}: {pass_hat_k}")



"""Shopify document loader implementation for the Arklex framework.

This module provides a specialized document loader for Shopify data, implementing
the base Loader interface. It includes functionality for loading product information
from Shopify's GraphQL API and processing the data into document chunks suitable
for further processing. The loader handles product descriptions, titles, and metadata,
making it easy to integrate Shopify product data into the framework's document
processing pipeline.

Key Components:
1. ShopifyLoader: Specialized loader for Shopify product data
   - load: Loads product data from Shopify's GraphQL API
   - chunk: Splits product documents into manageable chunks

Features:
- Integration with Shopify's GraphQL API
- Product data extraction and processing
- Metadata preservation during chunking
- Configurable text splitting parameters
- Support for product descriptions and titles

Usage:
    from arklex.utils.loaders.shopify import ShopifyLoader

    # Initialize the loader
    loader = ShopifyLoader()

    # Load product data
    products = loader.load()

    # Split into chunks
    chunks = loader.chunk(products)
"""

import json

import shopify
from langchain.text_splitter import RecursiveCharacterTextSplitter
from langchain_core.documents import Document

from arklex.utils.loaders.base import Loader


class ShopifyLoader(Loader):
    """Specialized loader for Shopify product data.

    This class implements the base Loader interface to provide functionality
    for loading and processing Shopify product data. It retrieves product
    information through Shopify's GraphQL API and processes it into document
    chunks suitable for further analysis or processing.

    The loader handles:
    - Product titles and descriptions
    - Product tags and metadata
    - Inventory information
    - Document chunking with metadata preservation

    Attributes:
        None

    Methods:
        load: Loads product data from Shopify's GraphQL API
        chunk: Splits product documents into manageable chunks
    """

    def __init__(self) -> None:
        """Initialize the ShopifyLoader instance."""

    def load(self) -> list[Document]:
        """Load product data from Shopify's GraphQL API.

        This function retrieves product information from Shopify's GraphQL API,
        including titles, descriptions, tags, and inventory data. It processes
        the data into Document objects suitable for further processing.

        The function:
        1. Executes a GraphQL query to fetch product data
        2. Processes the response into Document objects
        3. Preserves product metadata
        4. Returns a list of processed documents

        Returns:
            List[Document]: List of documents containing product information.
            Each document contains:
            - page_content: Product description
            - metadata: Product title

        Raises:
            shopify.ShopifyError: If there's an error accessing the Shopify API
            json.JSONDecodeError: If the API response cannot be parsed
        """
        docs = []
        response = shopify.GraphQL().execute("""
            {
                products(first: 23) {
                    edges {
                        node {
                            title
                            tags
                            description
                            totalInventory
                        }
                    }
                }
            }
            """)
        product_docs = json.loads(response)["data"]["products"]["edges"]
        for product_doc in product_docs:
            docs.append(
                Document(
                    page_content=product_doc["node"]["description"],
                    metadata={"title": product_doc["node"]["title"]},
                )
            )
        return docs

    def chunk(self, document_objs: list[Document]) -> list[Document]:
        """Split product documents into smaller chunks.

        This function splits product documents into smaller, more manageable chunks
        while preserving their metadata. It uses a text splitter to create chunks
        of appropriate size for processing.

        The chunking process:
        1. Initializes a text splitter with specified parameters
        2. Processes each product document
        3. Splits the content into chunks
        4. Preserves product metadata for each chunk
        5. Creates new Document objects for each chunk

        Args:
            document_objs (List[Document]): List of product documents to chunk.
                Each document should contain:
                - page_content: Product description
                - metadata: Product metadata

        Returns:
            List[Document]: List of chunked product documents.
            Each chunk preserves the original document's metadata.

        Note:
            The current implementation uses:
            - cl100k_base encoding
            - 200 token chunk size
            - 40 token chunk overlap
        """
        text_splitter = RecursiveCharacterTextSplitter.from_tiktoken_encoder(
            encoding_name="cl100k_base", chunk_size=200, chunk_overlap=40
        )
        docs = []
        langchain_docs = []
        for doc in document_objs:
            splitted_text = text_splitter.split_text(doc.page_content)
            for _i, txt in enumerate(splitted_text):
                docs.append(doc)
                langchain_docs.append(Document(page_content=txt, metadata=doc.metadata))
        return langchain_docs



from enum import Enum


class ResourceType(str, Enum):
    """Resource type enum."""

    TOOL = "tool"
    WORKER = "worker"
    AGENT = "agent"


class ToolCategory(str, Enum):
    """Tool category enum for organizing different types of tools."""

    GOOGLE_CALENDAR = "google-calendar"
    SHOPIFY = "shopify"
    HUBSPOT = "hubspot"
    TWILIO = "twilio"
    CUSTOM = "custom"


class WorkerCategory(str, Enum):
    """Worker category enum for organizing different types of workers."""

    WORKER = "worker"


class AgentCategory(str, Enum):
    """Agent category enum for organizing different types of agents."""

    OPENAI = "openai"


class Item(str, Enum):
    """Item enum for organizing different types of items."""


class ToolItem(Item):
    """Specific tool items organized by category."""

    # Google Calendar Tools
    GOOGLE_CREATE_EVENT = "google/calendar/create-event"
    GOOGLE_FREE_BUSY = "google/calendar/free-busy"

    # Shopify Tools
    SHOPIFY_FIND_USER_ID_BY_EMAIL = "shopify/find-user-id-by-email"
    SHOPIFY_GET_ORDER_DETAILS = "shopify/get-order-details"
    SHOPIFY_GET_PRODUCTS = "shopify/get-products"
    SHOPIFY_SEARCH_PRODUCTS = "shopify/search-products"
    SHOPIFY_CANCEL_ORDER = "shopify/cancel-order"
    SHOPIFY_CART_ADD_ITEMS = "shopify/cart-add-items"
    SHOPIFY_GET_CART = "shopify/get-cart"
    SHOPIFY_GET_USER_DETAILS_ADMIN = "shopify/get-user-details-admin"
    SHOPIFY_GET_WEB_PRODUCT = "shopify/get-web-product"
    SHOPIFY_RETURN_PRODUCTS = "shopify/return-products"

    # HubSpot Tools
    HUBSPOT_BOOK_MEETING = "hubspot/meeting/book-meeting"
    HUBSPOT_CHECK_AVAILABILITY = "hubspot/meeting/check-availability"
    HUBSPOT_CHECK_AVAILABLE = "hubspot/meeting/check-available"
    HUBSPOT_CREATE_MEETING = "hubspot/meeting/create-meeting"
    HUBSPOT_CREATE_TICKET = "hubspot/ticket/create-ticket"
    HUBSPOT_FIND_CONTACT_BY_EMAIL = "hubspot/contact/find-contact-by-email"
    HUBSPOT_FIND_OWNER_ID_BY_CONTACT_ID = "hubspot/contact/find-owner-id-by-contact-id"

    # Twilio Tools
    TWILIO_SMS_SEND_SMS = "twilio/sms/send-sms"
    TWILIO_SMS_SEND_PREDEFINED_SMS = "twilio/sms/send-predefined-sms"
    TWILIO_CALL_END_CALL = "twilio/call/end-call"
    TWILIO_CALL_VOICEMAIL = "twilio/call/voicemail"
    TWILIO_CALL_TRANSFER = "twilio/call/transfer"

    # Custom Tools
    HTTP_TOOL = "custom/http-tool"
    RETRIEVER = "custom/retriever"


class WorkerItem(Item):
    """Specific worker items organized by category."""

    # Workers
    MESSAGE_WORKER = "message-worker"
    FAISS_RAG_WORKER = "faiss-rag-worker"
    MILVUS_RAG_WORKER = "milvus-rag-worker"
    MULTIPLE_CHOICE_WORKER = "multiple-choice-worker"
    RAG_MESSAGE_WORKER = "rag-message-worker"
    SEARCH_WORKER = "search-worker"
    HUMAN_IN_THE_LOOP_WORKER = "human-in-the-loop-worker"
    OUTPUT_PROCESS_WORKER = "output-process-worker"


class AgentItem(Item):
    """Specific agent items organized by category."""

    OPENAI_AGENT = "openai-agent"
    OPENAI_REALTIME_VOICE_AGENT = "openai-realtime-voice-agent"



import logging
from datetime import datetime

import pytz
from hubspot import HubSpot

from arklex.env.tools.hubspot.base.entities import HubspotAuthTokens
from arklex.env.tools.tools import register_tool

logger = logging.getLogger(__name__)

description = "Schedule a meeting for the customer with the specific representative."

slots = [
    {
        "name": "first_name",
        "type": "str",
        "required": True,
        "description": "First name of the customer.",
    },
    {
        "name": "last_name",
        "type": "str",
        "required": True,
        "description": "Last name of the customer.",
    },
    {
        "name": "email",
        "type": "str",
        "required": True,
        "description": "Email of the customer.",
    },
    {
        "name": "start_time",
        "type": "str",
        "required": True,
        "description": f"The start time that the meeting will take place. The meeting's start time includes the hour, as the date alone is not sufficient. Datetime in ISO8601 format without timezone like 'YYYY-MM-DDTHH:MM:SS'. Eg. {datetime.now().isoformat()}.",
    },
    {
        "name": "duration",
        "type": "int",
        "enum": [15, 30, 60],
        "description": "The duration of the meeting in minutes.",
        "required": True,
    },
    {
        "name": "timezone",
        "type": "str",
        "enum": pytz.common_timezones,
        "description": "The timezone of the user. For example, 'America/New_York'.",
        "prompt": "Could you please provide your timezone or where are you now?",
        "required": True,
    },
]


@register_tool(description, slots)
def book_meeting(
    first_name: str,
    last_name: str,
    email: str,
    start_time: str,
    duration: int,
    timezone: str,
    auth: HubspotAuthTokens,
    **kwargs: dict[str, object],
) -> str:
    slug = kwargs.get("slug")
    api_client = HubSpot(access_token=auth["access_token"])

    logger.info(
        f"Booking a meeting for {first_name} {last_name} with {slug} at {start_time} for {duration} minutes in {timezone} timezone."
    )

    duration_ms = duration * 60 * 1000
    try:
        # Set the timezone for the start time
        tz = pytz.timezone(timezone)
        start_time_obj = datetime.fromisoformat(start_time)
        start_time_obj = tz.localize(start_time_obj)
        start_timestamp_ms = int(start_time_obj.timestamp() * 1000)
        logger.info(
            f"Start time: {start_time_obj}, start timestamp: {start_timestamp_ms}"
        )
    except ValueError:
        return "error: the start time is not in the correct format"

    res = api_client.api_request(
        {
            "path": "/scheduler/v3/meetings/meeting-links/book",
            "method": "POST",
            "body": {
                "slug": slug,
                "duration": duration_ms,
                "startTime": start_timestamp_ms,
                "email": email,
                "firstName": first_name,
                "lastName": last_name,
                "timezone": timezone,
                "locale": "en-us",
            },
        }
    )

    res = res.json()
    logger.info(f"Meeting book response: {res}")
    if res.get("status") == "error":
        return "error: " + res.get("message")
    return "The meeting has been booked successfully"



"""Data management utilities for task editor components.

This module contains the TaskDataManager class that handles task data operations
independently of UI components, making it easier to test and maintain.
"""

from typing import Any

from .protocols import TreeNodeProtocol, TreeProtocol


class TaskDataManager:
    """Manages task data operations independently of UI."""

    @staticmethod
    def extract_label_text(label: str | object) -> str:
        """Extract text from a label object.

        Args:
            label: The label object which may have different formats

        Returns:
            The extracted text string
        """
        if hasattr(label, "plain"):
            return label.plain
        return str(label)

    @staticmethod
    def build_tasks_from_tree(root: TreeNodeProtocol | None) -> list[dict[str, Any]]:
        """Build tasks list from tree structure.

        Args:
            root: The root node of the tree

        Returns:
            List of task dictionaries
        """
        if not root:
            return []

        updated_tasks = []
        for task_node in root.children:
            task_name = TaskDataManager.extract_label_text(task_node.label)
            task = {"name": task_name}
            steps = []

            for step_node in task_node.children:
                step_name = TaskDataManager.extract_label_text(step_node.label)
                steps.append(step_name)

            if steps:
                task["steps"] = steps
            updated_tasks.append(task)

        return updated_tasks

    @staticmethod
    def populate_tree_from_tasks(
        tree: TreeProtocol, tasks: list[dict[str, Any]]
    ) -> None:
        """Populate tree with task data.

        Args:
            tree: The tree widget to populate
            tasks: List of task dictionaries
        """
        if not tree.root:
            return

        tree.root.expand()
        tasks = tasks if tasks is not None else []

        for task in tasks:
            if tree.root:
                task_node = tree.root.add(task["name"])
                if "steps" in task and task["steps"]:
                    for step in task["steps"]:
                        if isinstance(step, dict):
                            step_text = step.get("description", str(step))
                        else:
                            step_text = str(step)
                        task_node.add_leaf(step_text)



"""Task graph formatter for Arklex framework.

Formats task definitions into graph structure with nodes, edges, and metadata.
Handles LLM-based intent generation and nested graph connectivity.
"""

from typing import Any

from arklex.utils.logging_utils import LogContext

log_context = LogContext(__name__)


class TaskGraphFormatter:
    """Formats task graphs.

    This class handles formatting of task graphs, ensuring they are properly
    structured and contain all required components.

    Attributes:
        _role (str): Role of the assistant
        _user_objective (str): User's objective
        _builder_objective (str): Builder's objective
        _domain (str): Domain of the task
        _intro (str): Introduction text
        _task_docs (List[Dict[str, Any]]): Task documentation
        _rag_docs (List[Dict[str, Any]]): RAG documentation
        _workers (List[Dict[str, Any]]): List of workers
    """

    # Default worker names - can be overridden in config
    DEFAULT_MESSAGE_WORKER = "MessageWorker"
    DEFAULT_RAG_WORKER = "FaissRAGWorker"
    DEFAULT_SEARCH_WORKER = "SearchWorker"
    DEFAULT_NESTED_GRAPH = "NestedGraph"
    DEFAULT_OUTPUT_PROCESS_WORKER = "OutputProcessWorker"

    def __init__(
        self,
        role: str = "",
        user_objective: str = "",
        builder_objective: str = "",
        domain: str = "",
        intro: str = "",
        task_docs: list[dict[str, Any]] | None = None,
        rag_docs: list[dict[str, Any]] | None = None,
        workers: list[dict[str, Any]] | None = None,
        tools: list[dict[str, Any]] | None = None,
        nluapi: str = "",
        slotfillapi: str = "",
        default_intent: str | None = None,
        default_weight: int = 1,
        default_pred: bool = False,
        default_definition: str = "",
        default_sample_utterances: list[str] | None = None,
        nodes: list[Any] | None = None,
        edges: list[Any] | None = None,
        allow_nested_graph: bool = True,
        model: object | None = None,
        settings: dict[str, Any] | None = None,
    ) -> None:
        """Initialize the TaskGraphFormatter.

        Args:
            role (str): Role of the assistant
            user_objective (str): User's objective
            builder_objective (str): Builder's objective
            domain (str): Domain of the task
            intro (str): Introduction text
            task_docs (Optional[List[Dict[str, Any]]]): Task documentation
            rag_docs (Optional[List[Dict[str, Any]]]): RAG documentation
            workers (Optional[List[Dict[str, Any]]]): List of workers
            tools (Optional[List[Dict[str, Any]]]): List of tools
            nluapi (str): NLU API
            slotfillapi (str): Slotfill API
            default_intent (Optional[str]): Default intent for edges
            default_weight (int): Default weight for edges
            default_pred (bool): Default pred value for edge attributes
            default_definition (str): Default definition for edge attributes
            default_sample_utterances (Optional[List[str]]): Default sample utterances for edge attributes
            nodes (Optional[List[Any]]): List of nodes
            edges (Optional[List[Any]]): List of edges
            allow_nested_graph (bool): Whether to allow nested graph generation
            model (Optional[Any]): Language model for intent generation
            settings (Optional[Dict[str, Any]]): Additional configuration settings
        """
        self._role = role
        self._user_objective = user_objective
        self._builder_objective = builder_objective
        self._domain = domain
        self._intro = intro
        self._task_docs = task_docs or []
        self._rag_docs = rag_docs or []
        self._workers = workers or []
        self._tools = tools or []
        self._nluapi = nluapi
        self._slotfillapi = slotfillapi
        self._default_intent = default_intent
        self._default_weight = default_weight
        self._default_pred = default_pred
        self._default_definition = default_definition
        self._default_sample_utterances = default_sample_utterances or []
        self._nodes = nodes
        self._edges = edges
        self._allow_nested_graph = allow_nested_graph
        self._model = model
        self._settings = settings

    def _find_worker_by_name(self, worker_name: str) -> dict[str, str]:
        """Get worker info from name, with fallback mappings.

        Args:
            worker_name (str): Name of the worker to find info for

        Returns:
            Dict[str, str]: Worker info with 'id' and 'name' keys
        """
        if self._workers:
            for worker in self._workers:
                if isinstance(worker, dict) and worker.get("name") == worker_name:
                    return {
                        "id": worker.get("id", worker_name.lower()),
                        "name": worker.get("name", worker_name),
                    }

        # Fallback mappings based on config
        fallback_workers = {
            "MessageWorker": {
                "id": "26bb6634-3bee-417d-ad75-23269ac17bc3",
                "name": "MessageWorker",
            },
            "FaissRAGWorker": {"id": "FaissRAGWorker", "name": "FaissRAGWorker"},
            "SearchWorker": {
                "id": "9c15af81-04b3-443e-be04-a3522124b905",
                "name": "SearchWorker",
            },
            "OutputProcessWorker": {
                "id": "OutputProcessWorker",
                "name": "OutputProcessWorker",
            },
        }
        return fallback_workers.get(
            worker_name, {"id": worker_name.lower(), "name": worker_name}
        )

    def format_task_graph(self, tasks: list[dict[str, Any]]) -> dict[str, Any]:
        """Format tasks into complete graph structure with nodes, edges, and metadata.

        Args:
            tasks (List[Dict[str, Any]]): List of task definitions to format

        Returns:
            Dict[str, Any]: Complete task graph with nodes, edges, and metadata
        """
        if self._nodes is not None and self._edges is not None:
            return {"nodes": self._nodes, "edges": self._edges}

        # Format nodes and edges
        nodes, node_lookup, all_task_node_ids = self._format_nodes(tasks)
        start_node_id = "0"  # Start node is always "0" in our current implementation
        edges, nested_graph_nodes = self._format_edges(
            tasks, node_lookup, all_task_node_ids, start_node_id
        )

        # Update NestedGraph node values with their target node IDs
        all_nodes = nodes + nested_graph_nodes
        for node_id, node_data in all_nodes:
            if node_data.get("resource", {}).get("name") == "NestedGraph":
                # Find the target node this NestedGraph connects to by looking at edges
                target_node_id = None
                for edge in edges:
                    if edge[0] == node_id:  # This edge starts from the NestedGraph node
                        # The target is the second element
                        potential_target = edge[1]
                        # Don't point to node "0" (start node)
                        if potential_target != "0":
                            target_node_id = potential_target
                            break

                if target_node_id:
                    node_data["attribute"]["value"] = target_node_id
                else:
                    # If no valid target found, point to the first non-start task node
                    task_node_ids = [
                        nid
                        for nid, ndata in all_nodes
                        if ndata.get("resource", {}).get("name")
                        in [
                            "MessageWorker",
                            "FaissRAGWorker",
                            "SearchWorker",
                            "OutputProcessWorker",
                        ]
                        and nid != "0"
                    ]
                    if task_node_ids:
                        node_data["attribute"]["value"] = task_node_ids[0]
                    else:
                        # Fallback: point to node "1" if it exists
                        node_data["attribute"]["value"] = "1"
            else:
                # If the value is a dict, replace with its description
                value = node_data.get("attribute", {}).get("value")
                if isinstance(value, dict):
                    desc = value.get("description")
                    if desc:
                        node_data["attribute"]["value"] = desc
                    else:
                        node_data["attribute"]["value"] = str(value)
                # If the value is a list or nested, flatten to string
                elif isinstance(value, list):
                    node_data["attribute"]["value"] = ", ".join(str(v) for v in value)

        # Add nested graph nodes to the main nodes list
        nodes.extend(nested_graph_nodes)

        reusable_tasks = {}  # Define reusable_tasks, can be passed in later

        graph = {
            "nodes": nodes,
            "edges": edges,
            "tasks": tasks,  # Include tasks for connectivity logic
            "role": self._role,
            "user_objective": self._user_objective,
            "builder_objective": self._builder_objective,
            "domain": self._domain,
            "intro": self._intro,
            "task_docs": self._task_docs,
            "rag_docs": self._rag_docs,
            "workers": self._workers,
            "tools": self._tools,
            "nluapi": self._nluapi,
            "slotfillapi": self._slotfillapi,
            "reusable_tasks": reusable_tasks,
            "settings": self._settings,
        }

        # Ensure nested graphs are connected correctly as sequential steps
        return self.ensure_nested_graph_connectivity(graph)

    def _format_nodes(self, tasks: list[dict]) -> tuple[list, dict, list]:
        """Create nodes for start, tasks, and steps with worker assignments.

        Args:
            tasks (List[Dict]): List of task definitions to create nodes for

        Returns:
            Tuple[List, Dict, List]: (nodes, node_lookup, all_task_node_ids)
                - nodes: List of formatted node data
                - node_lookup: Mapping of task identifiers to node IDs
                - all_task_node_ids: List of all task node IDs
        """
        nodes = []
        node_id_counter = 0
        node_lookup = {}
        all_task_node_ids = []

        # Create start node
        start_node_id = str(node_id_counter)
        message_worker_id = self._find_worker_by_name(self.DEFAULT_MESSAGE_WORKER)
        nodes.append(
            [
                start_node_id,
                {
                    "resource": {
                        "id": message_worker_id["id"],
                        "name": message_worker_id["name"],
                    },
                    "attribute": {
                        "value": "Hello! I'm here to assist you with any customer service inquiries.",
                        "task": "start message",
                        "directed": False,
                    },
                    "limit": 1,
                    "type": "start",
                },
            ]
        )
        node_id_counter += 1

        # First pass: Create all task and step nodes
        for task_idx, task in enumerate(tasks):
            # Use 'id' if available, otherwise generate one
            task_identifier = task.get("id", f"task_{task_idx}")

            # Ensure task has an id field
            if "id" not in task:
                task["id"] = task_identifier

            # Use the resource from the task if available, otherwise default to MessageWorker
            resource_name = self.DEFAULT_MESSAGE_WORKER
            if task.get("resource") and isinstance(task["resource"], dict):
                resource_name = task["resource"].get(
                    "name", self.DEFAULT_MESSAGE_WORKER
                )
            elif task.get("resource"):
                resource_name = str(task["resource"])

            # Handle nested graph resources with specific names
            if resource_name == self.DEFAULT_NESTED_GRAPH or (
                resource_name
                not in [
                    self.DEFAULT_MESSAGE_WORKER,
                    self.DEFAULT_RAG_WORKER,
                    self.DEFAULT_SEARCH_WORKER,
                    self.DEFAULT_OUTPUT_PROCESS_WORKER,
                ]
                and "workflow" in resource_name.lower()
            ):
                resource_info = {"id": "nested_graph", "name": "NestedGraph"}
            else:
                resource_info = self._find_worker_by_name(resource_name)

            task_node_id = str(node_id_counter)
            node_data = {
                "resource": {"id": resource_info["id"], "name": resource_info["name"]},
                "attribute": {
                    "value": task.get("description", ""),
                    "task": task.get("name", ""),
                    "directed": False,
                },
            }
            if task.get("type"):
                node_data["type"] = task.get("type")
            if task.get("limit"):
                node_data["limit"] = task.get("limit")

            nodes.append([task_node_id, node_data])
            node_lookup[task_identifier] = task_node_id
            all_task_node_ids.append(task_node_id)
            node_id_counter += 1

            # Ensure steps are properly broken down
            steps = task.get("steps", [])
            if not steps and task.get("description"):
                # If no steps defined, create a single step from the task description
                steps = [
                    {"description": task.get("description", ""), "step_id": "step_1"}
                ]
                task["steps"] = steps

            for step_idx, step in enumerate(steps):
                step_id = f"{task_identifier}_step{step_idx}"
                step_node_id = str(node_id_counter)

                # Use the resource from the step if available (from best practice manager)
                step_worker_name = self.DEFAULT_MESSAGE_WORKER
                if (
                    isinstance(step, dict)
                    and step.get("resource")
                    and isinstance(step["resource"], dict)
                ):
                    step_worker_name = step["resource"].get(
                        "name", self.DEFAULT_MESSAGE_WORKER
                    )
                elif isinstance(step, dict) and step.get("resource"):
                    step_worker_name = str(step["resource"])

                # Handle nested graph resources with specific names
                if step_worker_name == self.DEFAULT_NESTED_GRAPH or (
                    step_worker_name
                    not in [
                        self.DEFAULT_MESSAGE_WORKER,
                        self.DEFAULT_RAG_WORKER,
                        self.DEFAULT_SEARCH_WORKER,
                        self.DEFAULT_OUTPUT_PROCESS_WORKER,
                    ]
                    and "workflow" in step_worker_name.lower()
                ):
                    step_worker_info = {"id": "nested_graph", "name": "NestedGraph"}
                else:
                    step_worker_info = self._find_worker_by_name(step_worker_name)

                # Simplify step value to use simple string instead of complex nested structure
                if isinstance(step, dict):
                    # Check if step has a complex structure with task, description, step_id, etc.
                    if "task" in step and "description" in step and "step_id" in step:
                        # This is a complex step structure, extract just the description
                        step_value = step.get("description", "")
                    else:
                        # Simple step structure, use description or task
                        step_value = step.get("description", step.get("task", ""))
                elif isinstance(step, str):
                    step_value = step
                else:
                    step_value = str(step)

                nodes.append(
                    [
                        step_node_id,
                        {
                            "resource": {
                                "id": step_worker_info["id"],
                                "name": step_worker_info["name"],
                            },
                            "attribute": {
                                "value": step_value,
                                "task": task.get("name", ""),
                                "directed": False,
                            },
                        },
                    ]
                )
                node_lookup[step_id] = step_node_id
                node_id_counter += 1

        return nodes, node_lookup, all_task_node_ids

    def _create_edge_attributes(
        self,
        intent: str | None = None,
        weight: int = 1,
        pred: bool = False,
        definition: str = "",
        sample_utterances: list[str] | None = None,
    ) -> dict[str, Any]:
        """Create edge attributes with intent and metadata.

        Args:
            intent (Optional[str]): Intent for the edge, defaults to None
            weight (int): Edge weight, defaults to 1
            pred (bool): Prediction flag, defaults to False
            definition (str): Edge definition, defaults to empty string
            sample_utterances (Optional[List[str]]): Sample utterances for the edge

        Returns:
            Dict[str, Any]: Edge attributes dictionary with intent and metadata
        """
        return {
            "intent": intent,
            "attribute": {
                "weight": weight,
                "pred": pred,
                "definition": definition,
                "sample_utterances": sample_utterances or [],
            },
        }

    def _format_edges(
        self,
        tasks: list[dict[str, Any]],
        node_lookup: dict[str, str],
        all_task_node_ids: list[str],
        start_node_id: str,
    ) -> tuple[list[Any], list[Any]]:
        """Create edges between nodes with LLM-generated intents.

        Args:
            tasks (List[Dict[str, Any]]): List of task definitions
            node_lookup (Dict[str, str]): Mapping of task identifiers to node IDs
            all_task_node_ids (List[str]): List of all task node IDs
            start_node_id (str): ID of the start node

        Returns:
            Tuple[List[Any], List[Any]]: (edges, nested_graph_nodes)
                - edges: List of formatted edge data
                - nested_graph_nodes: List of nested graph nodes
        """
        edges = []
        nested_graph_nodes = []

        # Generate descriptive intents for main task edges using LLM
        for task in tasks:
            task_identifier = task.get("id", f"task_{tasks.index(task)}")
            task_node_id = node_lookup.get(task_identifier)

            if task_node_id:
                dependencies = task.get("dependencies", [])
                if not dependencies:
                    # This is a main task edge from the start node.
                    # Use the intent from the task, or generate a descriptive one
                    intent = task.get("intent", "")
                    if not intent:
                        # Generate a descriptive intent based on task name and description
                        task_name = task.get("name", "")
                        task_description = task.get("description", "")

                        # Create a prompt for intent generation
                        intent_prompt = f"""
                        Given a task with the following details, generate a user-facing intent that describes what the user wants to accomplish.
                        
                        Task Name: {task_name}
                        Task Description: {task_description}
                        
                        Generate a natural, user-facing intent that describes what the user is trying to achieve. 
                        Examples of good intents:
                        - "User inquires about purchasing options"
                        - "User wants to explore rental options"
                        - "User asks about delivery times and logistics"
                        - "User has technical support or troubleshooting queries"
                        - "User seeks information on robot capabilities and features"
                        - "User is interested in customization options"
                        - "User inquires about service and maintenance"
                        - "User wants to learn about new releases and updates"
                        
                        Intent: """

                        try:
                            # Use the model to generate intent if available
                            if self._model:
                                from langchain_core.messages import HumanMessage

                                response = self._model.invoke(
                                    [HumanMessage(content=intent_prompt)]
                                )
                                intent = response.content.strip().strip('"').strip("'")
                                # Clean up the response to ensure it's a valid intent
                                if not intent or intent.lower() in [
                                    "none",
                                    "null",
                                    "n/a",
                                ]:
                                    raise ValueError("Invalid intent generated")
                            else:
                                intent = "User inquires about purchasing options"  # default fallback
                        except Exception as e:
                            log_context.warning(
                                f"Failed to generate intent for task {task_name}: {e}"
                            )
                            intent = (
                                "User inquires about purchasing options"  # fallback
                            )

                    edges.append(
                        [
                            start_node_id,
                            task_node_id,
                            self._create_edge_attributes(intent=intent, pred=True),
                        ]
                    )
                else:
                    # This handles dependencies between tasks.
                    for dep in dependencies:
                        # Handle edge cases where dependency is None or not a string/dict
                        if dep is None:
                            log_context.warning("Skipping None dependency")
                            continue
                        elif isinstance(dep, str):
                            dep_id = dep
                        elif isinstance(dep, dict):
                            dep_id = dep.get("id")
                            if dep_id is None:
                                log_context.warning(
                                    "Skipping dependency dict without 'id' field"
                                )
                                continue
                        else:
                            log_context.warning(
                                f"Skipping invalid dependency type: {type(dep)}"
                            )
                            continue

                        # Find the source task from the list of tasks
                        source_task = next(
                            (t for t in tasks if t.get("id") == dep_id), None
                        )

                        if source_task:
                            source_steps = source_task.get("steps", [])
                            if source_steps:
                                # Dependency should come from the last step of the source task
                                last_step_identifier = (
                                    f"{dep_id}_step{len(source_steps) - 1}"
                                )
                                source_node_id = node_lookup.get(last_step_identifier)
                            else:
                                # If no steps, dependency is from the task node itself
                                source_node_id = node_lookup.get(dep_id)
                        else:
                            # Fallback to original logic if source task not found
                            source_node_id = node_lookup.get(dep_id)

                        if source_node_id:
                            edges.append(
                                [
                                    source_node_id,
                                    task_node_id,
                                    self._create_edge_attributes(),
                                ]
                            )
                        else:
                            log_context.warning(
                                f"Could not find source node for dependency '{dep_id}'"
                            )

        # Steps
        for task in tasks:
            task_identifier = task.get("id", f"task_{tasks.index(task)}")
            task_node_id = node_lookup.get(task_identifier)

            if task_node_id:
                steps = task.get("steps", [])
                if steps:
                    first_step_id = f"{task_identifier}_step0"
                    if first_step_id in node_lookup:
                        edges.append(
                            [
                                task_node_id,
                                node_lookup[first_step_id],
                                self._create_edge_attributes(intent=None),
                            ]
                        )
                        for i in range(len(steps) - 1):
                            current_step_id = f"{task_identifier}_step{i}"
                            next_step_id = f"{task_identifier}_step{i + 1}"
                            if (
                                current_step_id in node_lookup
                                and next_step_id in node_lookup
                            ):
                                edges.append(
                                    [
                                        node_lookup[current_step_id],
                                        node_lookup[next_step_id],
                                        self._create_edge_attributes(intent=None),
                                    ]
                                )

        return edges, nested_graph_nodes

    def ensure_nested_graph_connectivity(self, graph: dict[str, Any]) -> dict[str, Any]:
        """Ensures that all nested graph nodes are properly connected as sequential steps.

        This method finds all 'NestedGraph' nodes and connects them to the
        next step within their own task, preventing incorrect fan-out connections
        to leaf nodes.

        Args:
            graph (Dict[str, Any]): The graph structure.

        Returns:
            Dict[str, Any]: The graph with corrected nested graph connectivity.
        """
        nodes = graph.get("nodes", [])
        edges = graph.get("edges", [])
        tasks = graph.get("tasks", [])
        node_lookup = {node_data[0]: node_data[1] for node_data in nodes}
        # Create a reverse lookup from node ID to task_id
        node_to_task_map = {}
        for task in tasks:
            task_id = task.get("id")
            for i in range(len(task.get("steps", []))):
                step_node_id = f"{task_id}_step{i}"
                node_to_task_map[step_node_id] = task_id

        nested_graph_nodes = [
            (node_id, node_data)
            for node_id, node_data in node_lookup.items()
            if node_data.get("resource", {}).get("name") == self.DEFAULT_NESTED_GRAPH
        ]

        for ng_node_id, ng_node_data in nested_graph_nodes:
            # Find which task this nested_graph node belongs to
            task_id = node_to_task_map.get(ng_node_id)
            if not task_id:
                continue

            # Find the corresponding task and the index of the nested_graph step
            task = next((t for t in tasks if t.get("id") == task_id), None)
            if not task:
                continue

            steps = task.get("steps", [])
            step_index = -1
            for i, _step in enumerate(steps):
                step_node_id = f"{task_id}_step{i}"
                if step_node_id == ng_node_id:
                    step_index = i
                    break

            # If it's not the last step, connect it to the next one
            if 0 <= step_index < len(steps) - 1:
                next_step_node_id = f"{task_id}_step{step_index + 1}"

                # Set the value attribute - create attribute dict if it doesn't exist
                ng_node_data.setdefault("attribute", {})["value"] = next_step_node_id

                # Create the edge
                edges.append(
                    [
                        ng_node_id,
                        next_step_node_id,
                        self._create_edge_attributes(
                            definition="Continue to next step from nested graph"
                        ),
                    ]
                )

        graph["edges"] = edges
        return graph



"""FAISS RAG worker implementation for the Arklex framework.

This module provides a specialized worker for handling Retrieval-Augmented Generation (RAG)
tasks using FAISS for efficient similarity search. It implements a worker that can answer
user questions based on internal documentation, including policies, FAQs, and product
information. The worker supports both streaming and non-streaming responses, using a state
graph to manage the workflow of document retrieval and response generation.
"""

from typing import Any

from arklex.env.tools.RAG.retrievers.faiss_retriever import RetrieveEngine
from arklex.env.tools.utils import ToolGenerator
from arklex.env.workers.base.base_worker import BaseWorker
from arklex.env.workers.faiss_rag.entities import (
    FaissRAGWorkerData,
    FaissRAGWorkerOutput,
)
from arklex.orchestrator.entities.orchestrator_state_entities import (
    OrchestratorState,
    StatusEnum,
)
from arklex.types.stream_types import StreamType
from arklex.utils.logging_utils import LogContext

log_context = LogContext(__name__)


class FaissRAGWorker(BaseWorker):
    description: str = "Answer the user's questions based on the company's internal documentations (unstructured text data), such as the policies, FAQs, and product information"

    def __init__(self) -> None:
        super().__init__()

    def init_worker_data(
        self, orch_state: OrchestratorState, node_specific_data: dict[str, Any]
    ) -> None:
        self.orch_state = orch_state
        self.faiss_rag_worker_data: FaissRAGWorkerData = FaissRAGWorkerData(
            **node_specific_data,
        )

    def _execute(self) -> FaissRAGWorkerOutput:
        retrieved_text = RetrieveEngine.faiss_retrieve(self.orch_state)
        self.orch_state.message_flow = retrieved_text
        if self.orch_state.stream_type != StreamType.NON_STREAM:
            response = ToolGenerator.stream_context_generate(self.orch_state)
        else:
            response = ToolGenerator.context_generate(self.orch_state)

        return FaissRAGWorkerOutput(
            response=response,
            status=StatusEnum.COMPLETE,
        )



from typing import Any

from benchmark.tau_bench.envs.tool import Tool


class FindUserIdByNameZip(Tool):
    @staticmethod
    def invoke(data: dict[str, Any], first_name: str, last_name: str, zip: str) -> str:
        users = data["users"]
        for user_id, profile in users.items():
            if (
                profile["name"]["first_name"].lower() == first_name.lower()
                and profile["name"]["last_name"].lower() == last_name.lower()
                and profile["address"]["zip"] == zip
            ):
                return user_id
        raise Exception("Error: user not found")

    @staticmethod
    def get_info() -> dict[str, Any]:
        return {
            "type": "function",
            "function": {
                "name": "find_user_id_by_name_zip",
                "description": (
                    "Find user id by first name, last name, and zip code. If the user is not found, the function "
                    "will return an error message. By default, find user id by email, and only call this function "
                    "if the user is not found by email or cannot remember email."
                ),
                "parameters": {
                    "type": "object",
                    "properties": {
                        "first_name": {
                            "type": "string",
                            "description": "The first name of the customer, such as 'John'.",
                        },
                        "last_name": {
                            "type": "string",
                            "description": "The last name of the customer, such as 'Doe'.",
                        },
                        "zip": {
                            "type": "string",
                            "description": "The zip code of the customer, such as '12345'.",
                        },
                    },
                    "required": ["first_name", "last_name", "zip"],
                },
            },
        }



"""Graph state management for the Arklex framework.

This module defines the data structures and models for managing the state of conversation
graphs and task execution flows. It includes classes for representing bot configurations,
message states, task status, and various components of the conversation graph.
The module provides comprehensive state management for tracking conversation progress,
resource records, and orchestrator responses throughout the system's operation.

The module is organized into several key components:
1. Bot Configuration: Classes for managing bot and LLM settings
2. Message Handling: Classes for processing and managing conversation messages
3. Task Status: Classes for tracking task execution and timing
4. Graph Structure: Classes for managing the conversation graph and node states
5. Memory Management: Classes for maintaining conversation history and trajectories

Key Features:
- Comprehensive state management for conversation flows
- Support for multiple bot configurations and LLM providers
- Flexible message processing and handling
- Task execution tracking and timing
- Graph-based conversation structure
- Memory management for conversation history
"""

import uuid
from enum import Enum
from typing import Any

from pydantic import BaseModel, Field

from arklex.memory.entities.memory_entities import ResourceRecord
from arklex.types.stream_types import StreamType
from arklex.utils.llm_config import LLMConfig


# Bot-related classes
class BotConfig(BaseModel):
    """Configuration for bot settings.

    This class defines the overall configuration for a bot instance, including its
    identity, version, language settings, and language model configuration.

    The class provides:
    1. Bot identity and versioning
    2. Language and settings
    3. LLM configuration integration
    4. Type-safe configuration management

    Attributes:
        bot_id (str): Unique identifier for the bot.
        version (str): Version number of the bot.
        language (str): Primary language of the bot.
        llm_config (LLMConfig): Language model configuration for the bot.
    """

    bot_id: str = Field(default="")
    version: str = Field(default="")
    language: str = Field(default="EN")
    llm_config: LLMConfig = Field(default_factory=LLMConfig)


### Message-related classes
class ConvoMessage(BaseModel):
    """Represents a conversation message with history.

    This class encapsulates a single message in a conversation, including both
    the current message content and the conversation history leading up to it.

    The class provides:
    1. Message content management
    2. Conversation history tracking
    3. Type-safe message handling

    Attributes:
        history (str): Previous conversation history or its summarization.
        message (str): The current message content.
    """

    history: str  # it could be the whole original message or the summarization of the previous conversation from memory module
    message: str


### Task status-related classes
class StatusEnum(str, Enum):
    """Enumeration of possible task statuses.

    This enum defines the possible states that a task can be in during its execution.
    It's used throughout the system to track the progress and state of various operations.

    The enum provides:
    1. Clear status definitions
    2. Type-safe status handling
    3. Consistent status tracking

    Values:
        COMPLETE: Task has been completed successfully.
        INCOMPLETE: Task is not yet complete.
        STAY: Task should remain in its current state.
    """

    COMPLETE: str = "complete"
    INCOMPLETE: str = "incomplete"
    STAY: str = "stay"


class Timing(BaseModel):
    """Timing information for task execution.

    This class tracks timing information for various aspects of task execution,
    particularly focusing on task graph processing times.

    The class provides:
    1. Task timing tracking
    2. Graph processing time measurement
    3. Type-safe timing management

    Attributes:
        taskgraph (Optional[float]): Time taken for task graph processing.
    """

    taskgraph: float | None = None


class Metadata(BaseModel):
    """Metadata for tracking conversation and execution.

    This class maintains metadata about the conversation session, including
    identifiers, turn tracking, and timing information.

    The class provides:
    1. Session identification
    2. Turn tracking
    3. Timing information
    4. Human-in-the-loop status
    5. Type-safe metadata management

    Attributes:
        chat_id (str): Unique identifier for the chat session.
        turn_id (int): Current turn number in the conversation.
        hitl (Optional[str]): Human-in-the-loop intervention status.
        timing (Timing): Timing information for the session.
        attempts (Optional[int]): Number of attempts for HITL MC logic.
    """

    chat_id: str = Field(default_factory=lambda: str(uuid.uuid4()))
    turn_id: int = 0
    hitl: str | None = Field(default=None)
    timing: Timing = Field(default_factory=Timing)
    attempts: int | None = None


class OrchestratorResp(BaseModel):
    """Response from the orchestrator.

    This class represents the response generated by the orchestrator,
    including the answer, parameters, and any additional information
    needed for the conversation flow.

    The class provides:
    1. Answer management
    2. Parameter handling
    3. Human-in-the-loop support
    4. Choice list management
    5. Type-safe response handling

    Attributes:
        answer (str): The response answer.
        parameters (Dict[str, Any]): Additional parameters.
        human_in_the_loop (Optional[str]): Human intervention status.
        choice_list (Optional[List[str]]): List of available choices.
    """

    answer: str = Field(default="")
    parameters: dict[str, Any] = Field(default_factory=dict)
    human_in_the_loop: str | None = Field(default=None)
    choice_list: list[str] | None = Field(default=[])


class OrchestratorState(BaseModel):
    """State management for message processing.

    This class manages the complete state of a message as it moves through the system,
    including configuration, processing trajectory, and response generation.

    The class provides:
    1. System configuration management
    2. Bot configuration handling
    3. Message processing tracking
    4. Response generation
    5. Status tracking
    6. Slot management
    7. Streaming support
    8. Type-safe state management

    Attributes:
        sys_instruct (str): System instructions for message processing.
        bot_config (BotConfig): Configuration for the bot.
        metadata (Metadata): Session metadata.
        user_message (ConvoMessage): User's input message.
        message_flow (str): Flow of messages between nodes.
        function_calling_trajectory (List[Dict[str, Any]]): History of function calls.
        trajectory (List[List[ResourceRecord]]): Processing trajectory.
        relevant_records (Optional[List[ResourceRecord]]): Relevant resource records.
        stream_type (str): Type of streaming being used.
        message_queue (Any): Queue for streaming messages.
    """

    # system-level configuration
    sys_instruct: str = Field(default="")
    bot_config: BotConfig | None = Field(default=None)
    metadata: Metadata | None = Field(default=None)
    # execution fields
    user_message: ConvoMessage | None = Field(default=None)
    message_flow: str = Field(default="")
    # record history
    function_calling_trajectory: list[dict[str, Any]] | None = Field(default=None)
    trajectory: list[list[ResourceRecord]] | None = Field(default=None)
    relevant_records: list[ResourceRecord] | None = Field(default=None)
    # streaming
    stream_type: StreamType | None = Field(default=StreamType.NON_STREAM)
    message_queue: Any = Field(exclude=True, default=None)



"""ChatGPT and LLM utility functions for evaluation in the Arklex framework.

This module provides utility functions for interacting with various language models (OpenAI,
Anthropic, Gemini) during evaluation. It includes functionality for client creation,
message handling, conversation history management, and goal generation. The module supports
multiple LLM providers, conversation role flipping, history formatting, and chatbot querying
with parameter management.
"""

import json
import os
import random
from typing import Any

import anthropic
import requests
from dotenv import load_dotenv
from google.generativeai import GenerativeModel
from openai import OpenAI

from arklex.utils.model_config import MODEL
from arklex.utils.provider_utils import get_api_key_for_provider

load_dotenv()


def create_client() -> OpenAI | anthropic.Anthropic | GenerativeModel:
    """Create a client for interacting with language models.

    This function creates and returns a client for the configured language model provider
    (OpenAI, Anthropic, or Gemini). It handles API key configuration and organization
    settings.

    Returns:
        Union[OpenAI, anthropic.Anthropic, GenerativeModel]: A client instance for the configured LLM provider.

    Raises:
        KeyError: If required environment variables are not set.
        ValueError: If the configured provider is not supported.
    """
    try:
        org_key: str | None = os.environ["OPENAI_ORG_ID"]
    except KeyError:
        org_key = None

    provider = MODEL["llm_provider"]

    if not provider:
        raise ValueError(
            "llm_provider must be explicitly specified in MODEL configuration"
        )

    if provider == "openai":
        client: OpenAI = OpenAI(
            api_key=get_api_key_for_provider("openai"),
            organization=org_key,
        )
        return client

    elif provider == "google":
        # Set the API key for Google Generative AI
        import google.generativeai as genai

        genai.configure(api_key=get_api_key_for_provider("google"))
        # Use the model from MODEL configuration instead of hardcoding
        model_name = MODEL.get("model_type_or_path", "gemini-1.5-flash")
        client = GenerativeModel(model_name)
        return client

    elif provider == "anthropic":
        client: anthropic.Anthropic = anthropic.Anthropic(
            api_key=get_api_key_for_provider("anthropic")
        )
        return client

    else:
        raise ValueError(f"Unsupported LLM provider: {provider}")


def chatgpt_chatbot(
    messages: list[dict[str, str]],
    client: OpenAI | anthropic.Anthropic | GenerativeModel,
    model: str = None,
) -> str:
    """Send messages to a language model and get a response.

    This function sends a list of messages to the specified language model and returns
    its response. It handles different providers (OpenAI, Anthropic, Gemini) and their specific
    API requirements.

    Args:
        messages (List[Dict[str, str]]): List of message dictionaries with 'role' and 'content'.
        client (Union[OpenAI, anthropic.Anthropic, GenerativeModel]): The LLM client to use.
        model (str, optional): The model to use. Defaults to MODEL["model_type_or_path"].

    Returns:
        str: The model's response text.
    """
    # Get model from MODEL config if not provided
    if model is None:
        model = MODEL["model_type_or_path"]

    provider = MODEL["llm_provider"]

    if not provider:
        raise ValueError(
            "llm_provider must be explicitly specified in MODEL configuration"
        )

    # Ensure model is not empty
    if not model:
        raise ValueError(
            "Model parameter cannot be empty. Please check MODEL configuration."
        )

    if provider == "openai":
        answer: str = (
            client.chat.completions.create(
                model=model, messages=messages, temperature=0.1
            )
            .choices[0]
            .message.content.strip()
        )
    elif provider == "anthropic":
        kwargs: dict[str, Any] = {
            "model": model,
            "messages": messages if messages[0]["role"] != "system" else [messages[1]],
            "temperature": 0.1,
            "max_tokens": 1024,
            **(
                {"system": messages[0]["content"]}
                if messages[0]["role"] == "system"
                else {}
            ),
        }
        answer: str = client.messages.create(**kwargs).content[0].text.strip()
    elif provider == "google":
        # Convert messages to Gemini format
        gemini_messages = _convert_messages_to_gemini_format(messages)

        response = client.generate_content(
            gemini_messages,
            generation_config={"temperature": 0.1, "max_output_tokens": 1024},
        )
        answer: str = response.text.strip()
    else:
        raise ValueError(f"Unsupported LLM provider: {provider}")

    return answer


def _convert_messages_to_gemini_format(
    messages: list[dict[str, str]],
) -> list[dict[str, Any]]:
    """Convert OpenAI/Anthropic message format to Gemini format.

    This function converts a list of messages from OpenAI/Anthropic format to
    Gemini's expected format. System messages are prepended to the first user message
    since Gemini doesn't support system messages directly.

    Args:
        messages (List[Dict[str, str]]): List of messages in OpenAI/Anthropic format.

    Returns:
        List[Dict[str, Any]]: List of messages in Gemini format.
    """
    gemini_messages = []
    system_content = ""

    # Extract system message if present
    if messages and messages[0]["role"] == "system":
        system_content = messages[0]["content"]
        messages = messages[1:]  # Remove system message from processing

    # Convert remaining messages
    for msg in messages:
        if msg["role"] == "user":
            gemini_messages.append(
                {"role": "user", "parts": [{"text": msg["content"]}]}
            )
        elif msg["role"] == "assistant":
            gemini_messages.append(
                {"role": "model", "parts": [{"text": msg["content"]}]}
            )

    # Prepend system content to first user message if it exists
    if system_content and gemini_messages and gemini_messages[0]["role"] == "user":
        gemini_messages[0]["parts"][0]["text"] = (
            f"{system_content}\n\n{gemini_messages[0]['parts'][0]['text']}"
        )

    return gemini_messages


# flip roles in convo history, only keep role and content
def flip_hist_content_only(hist: list[dict[str, Any]]) -> list[dict[str, str]]:
    """Flip roles in conversation history, keeping only role and content.

    This function takes a conversation history and flips the roles (user becomes assistant
    and vice versa), while keeping only the 'role' and 'content' fields. System messages
    are removed.

    Args:
        hist (List[Dict[str, Any]]): The conversation history to process.

    Returns:
        List[Dict[str, str]]: The processed conversation history with flipped roles.
    """
    new_hist: list[dict[str, str]] = []
    for turn in hist:
        if turn["role"] == "system":
            continue
        elif turn["role"] == "user":
            new_hist.append({"role": "assistant", "content": turn["content"]})
        else:
            new_hist.append({"role": "user", "content": turn["content"]})
    return new_hist


# flip roles in convo history, keep all other keys the same
def flip_hist(hist: list[dict[str, Any]]) -> list[dict[str, Any]]:
    """Flip roles in conversation history, preserving all fields.

    This function takes a conversation history and flips the roles (user becomes assistant
    and vice versa), while preserving all other fields in the message dictionaries.
    System messages are removed.

    Args:
        hist (List[Dict[str, Any]]): The conversation history to process.

    Returns:
        List[Dict[str, Any]]: The processed conversation history with flipped roles.
    """
    new_hist: list[dict[str, Any]] = []
    for turn in hist.copy():
        if "role" not in turn:
            new_hist.append(turn)
        elif turn["role"] == "system":
            continue
        elif turn["role"] == "user":
            turn["role"] = "assistant"
            new_hist.append(turn)
        else:
            turn["role"] = "user"
            new_hist.append(turn)
    return new_hist


def query_chatbot(
    model_api: str,
    history: list[dict[str, Any]],
    params: dict[str, Any],
    env_config: dict[str, Any],
) -> dict[str, Any]:
    """Query a chatbot API with conversation history and parameters.

    This function sends a request to a chatbot API with the provided conversation history,
    parameters, and environment configuration. It flips the roles in the history before
    sending the request.

    Args:
        model_api (str): The URL of the chatbot API.
        history (List[Dict[str, Any]]): The conversation history.
        params (Dict[str, Any]): Additional parameters for the request.
        env_config (Dict[str, Any]): Environment configuration including workers and tools.

    Returns:
        Dict[str, Any]: The API response as a dictionary.
    """
    history = flip_hist_content_only(history)
    data: dict[str, Any] = {
        "history": history,
        "parameters": params,
        "workers": env_config["workers"],
        "tools": env_config["tools"],
    }
    data_str: str = json.dumps(data)
    response = requests.post(
        model_api, headers={"Content-Type": "application/json"}, data=data_str
    )
    return response.json()


def format_chat_history_str(chat_history: list[dict[str, str]]) -> str:
    """Format chat history into a single string.

    This function takes a list of chat messages and formats them into a single string,
    with each message prefixed by its role in uppercase.

    Args:
        chat_history (List[Dict[str, str]]): List of chat messages with 'role' and 'content'.

    Returns:
        str: The formatted chat history string.
    """
    formatted_hist: str = ""
    for turn in chat_history:
        formatted_hist += turn["role"].upper() + ": " + turn["content"] + " "
    return formatted_hist.strip()


# filter prompts out of bot utterances
def filter_convo(
    convo: list[dict[str, Any]], delim: str = "\n", filter_turns: bool = True
) -> list[dict[str, Any]]:
    """Filter and process a conversation.

    This function processes a conversation by:
    1. Skipping the first two turns
    2. Optionally filtering out turns without a 'role' field
    3. For assistant messages, keeping them as is
    4. For user messages, truncating content at the specified delimiter

    Args:
        convo (List[Dict[str, Any]]): The conversation to process.
        delim (str, optional): Delimiter for truncating user messages. Defaults to "\n".
        filter_turns (bool, optional): Whether to filter turns without roles. Defaults to True.

    Returns:
        List[Dict[str, Any]]: The processed conversation.
    """
    filtered_convo: list[dict[str, Any]] = []
    for i, turn in enumerate(convo):
        if i <= 1 or "role" not in turn and filter_turns:
            continue
        elif "role" not in turn or turn["role"] == "assistant":
            filtered_convo.append(turn)
        else:
            idx: int = turn["content"].find(delim)
            new_turn: dict[str, Any] = {}
            for key in turn:
                if key == "content":
                    if idx == -1:
                        new_turn[key] = turn[key]
                    else:
                        new_turn[key] = turn[key][:idx]
                else:
                    new_turn[key] = turn[key]
            filtered_convo.append(new_turn)
    return filtered_convo


def adjust_goal(doc_content: str, goal: str) -> str:
    """Adjust a goal based on document content.

    This function uses a language model to adjust a goal by replacing specific product
    mentions with products from the provided document content. If no specific products
    are mentioned in the goal, it returns the original goal unchanged.

    Args:
        doc_content (str): The document content to reference for product replacements.
        goal (str): The original goal to adjust.

    Returns:
        str: The adjusted goal.
    """
    message: str = f"Pretend you have the following goal in the mind. If the goal including some specific product, such as floss, mug, iphone, etc., then please replace it with the product from the following document content. Otherwise, don't need to change it and just return the original goal. The document content is as follows:\n{doc_content}\n\nThe original goal is as follows:\n{goal}\n\nOnly give the answer to the question in your response."

    client = create_client()
    return chatgpt_chatbot(
        [{"role": "user", "content": message}],
        client,
        model=MODEL["model_type_or_path"],
    )


def generate_goal(
    doc_content: str, client: OpenAI | anthropic.Anthropic | GenerativeModel
) -> str:
    """Generate a goal based on document content.

    This function uses a language model to generate a goal or information request
    based on the provided document content. The goal is generated in first person
    and represents what a user might want to achieve when interacting with a chatbot
    about the document's content.

    Args:
        doc_content (str): The document content to base the goal on.
        client (Union[OpenAI, anthropic.Anthropic, GenerativeModel]): The LLM client to use.

    Returns:
        str: The generated goal.
    """
    message: str = f"Pretend you have just read the following website:\n{doc_content}\nThis website also has a chatbot. What is some information you want to get from this chatbot or a goal you might have when chatting with this chatbot based on the website content? Answer the question in the first person. Only give the answer to the question in your response."

    return chatgpt_chatbot(
        [{"role": "user", "content": message}],
        client,
        model=MODEL["model_type_or_path"],
    )


def generate_goals(
    documents: list[dict[str, str]],
    params: dict[str, Any],
    client: OpenAI | anthropic.Anthropic | GenerativeModel,
) -> list[str]:
    """Generate multiple goals based on a collection of documents.

    This function generates a specified number of goals by randomly selecting documents
    and using the language model to generate goals based on their content.

    Args:
        documents (List[Dict[str, str]]): List of documents with their content.
        params (Dict[str, Any]): Parameters including 'num_goals' to generate.
        client (Union[OpenAI, anthropic.Anthropic, GenerativeModel]): The LLM client to use.

    Returns:
        List[str]: List of generated goals.
    """
    goals: list[str] = []
    for _ in range(params["num_goals"]):
        doc: dict[str, str] = random.choice(documents)
        goals.append(generate_goal(doc["content"], client))
    return goals


def _print_goals(goals: list[str]) -> None:
    print(goals)


def main() -> None:
    from get_documents import filter_documents, get_all_documents

    documents = get_all_documents()
    documents = filter_documents(documents)
    client = create_client()
    params = {"num_goals": 1}
    _print_goals(generate_goals(documents, params, client))


if __name__ == "__main__":
    main()



import enum
import json
import re
from typing import Any, TypeVar

from pydantic import BaseModel, Field

from benchmark.tau_bench.model_utils.api.types import PartialObj

T = TypeVar("T", bound=BaseModel)


class InputType(enum.Enum):
    CHAT = "chat"
    COMPLETION = "completion"


def display_choices(choices: list[str]) -> tuple[str, dict[str, int]]:
    choice_displays: list[str] = []
    decode_map: dict[str, int] = {}
    for i, choice in enumerate(choices):
        label: str = index_to_alpha(i)
        choice_display: str = f"{label}. {choice}"
        choice_displays.append(choice_display)
        decode_map[label] = i
    return "\n".join(choice_displays), decode_map


def index_to_alpha(index: int) -> str:
    alpha: str = ""
    while index >= 0:
        alpha = chr(index % 26 + ord("A")) + alpha
        index = index // 26 - 1
    return alpha


def type_to_json_schema_string(typ: type[T]) -> str:
    json_schema: dict[str, Any] = typ.model_json_schema()
    return json.dumps(json_schema, indent=4)


def optionalize_type(typ: type[T]) -> type[T]:
    class OptionalModel(typ): ...

    new_fields: dict[str, Any] = {}
    for name, field in OptionalModel.model_fields.items():
        new_fields[name] = Field(default=None, annotation=field.annotation | None)
    OptionalModel.model_fields = new_fields
    OptionalModel.__name__ = typ.__name__
    return OptionalModel


def json_response_to_obj_or_partial_obj(
    response: dict[str, Any], typ: type[T] | dict[str, Any]
) -> T | PartialObj | dict[str, Any]:
    if isinstance(typ, dict):
        return response
    else:
        required_field_names: list[str] = [
            name for name, field in typ.model_fields.items() if field.is_required()
        ]
        for name in required_field_names:
            if name not in response or response[name] is None:
                return response
        return typ.model_validate(response)


def clean_top_level_keys(d: dict[str, Any]) -> dict[str, Any]:
    new_d: dict[str, Any] = {}
    for k, v in d.items():
        new_d[k.strip()] = v
    return new_d


def parse_json_or_json_markdown(text: str) -> dict[str, Any]:
    def parse(s: str) -> dict[str, Any] | None:
        try:
            return json.loads(s)
        except json.decoder.JSONDecodeError:
            return None

    # pass #1: try to parse as json
    parsed: dict[str, Any] | None = parse(text)
    if parsed is not None:
        return parsed

    # pass #2: try to parse as json markdown
    stripped: str = text.strip()
    if stripped.startswith("```json"):
        stripped = stripped[len("```json") :].strip()
    if stripped.endswith("```"):
        stripped = stripped[: -len("```")].strip()
    parsed = parse(stripped)
    if parsed is not None:
        return parsed

    # pass #3: try to parse an arbitrary md block
    pattern: str = r"```(?:\w+\n)?(.*?)```"
    match: re.Match[str] | None = re.search(pattern, text, re.DOTALL)
    if match:
        content: str = match.group(1).strip()
        parsed = parse(content)
        if parsed is not None:
            return parsed

    # pass #4: try to parse arbitrary sections as json
    lines: list[str] = text.split("\n")
    seen: set[tuple[int, int]] = set()
    for i in range(len(lines)):
        for j in range(i + 1, len(lines) + 1):
            if i < j and (i, j) not in seen:
                seen.add((i, j))
                content = "\n".join(lines[i:j])
                parsed = parse(content)
                if parsed is not None:
                    return parsed
    raise ValueError("Could not parse JSON or JSON markdown")


def longest_valid_string(s: str, options: list[str]) -> str | None:
    longest: int = 0
    longest_str: str | None = None
    options_set: set[str] = set(options)
    for i in range(len(s)):
        if s[: i + 1] in options_set and i + 1 > longest:
            longest = i + 1
            longest_str = s[: i + 1]
    return longest_str


def try_classify_recover(s: str, decode_map: dict[str, int]) -> str | None:
    lvs: str | None = longest_valid_string(s, list(decode_map.keys()))
    if lvs is not None and lvs in decode_map:
        return lvs
    for k, v in decode_map.items():
        if s == v:
            return k
    return None


def approx_num_tokens(text: str) -> int:
    return len(text) // 4


def add_md_close_tag(prompt: str) -> str:
    return f"{prompt}\n```"


def add_md_tag(prompt: str) -> str:
    return f"```json\n{prompt}\n```"



from collections.abc import Callable, Iterable
from concurrent.futures import ThreadPoolExecutor
from typing import TypeVar

T = TypeVar("T")
U = TypeVar("U")


def map(
    func: Callable[[T], U],
    iterable: Iterable[T],
    max_concurrency: int | None = None,
    use_tqdm: bool = False,
) -> list[U]:
    assert max_concurrency is None or max_concurrency > 0
    with ThreadPoolExecutor(max_workers=max_concurrency) as executor:
        if use_tqdm:
            from tqdm import tqdm

            return list(tqdm(executor.map(func, iterable), total=len(iterable)))
        return list(executor.map(func, iterable))



"""Entities for memory management.

This module defines the entities used in the memory management system,
including ResourceRecord and Memory.

Key Components:
- ResourceRecord: Represents a record of resource usage and processing.
- Memory: Represents the memory of the conversation system.
"""

from typing import Any

from pydantic import BaseModel, Field


class ResourceRecord(BaseModel):
    """Record of resource usage and processing.

    This class maintains a record of how resources are used during task execution,
    including input/output data, processing steps, and intent information.

    The class provides:
    1. Resource usage tracking
    2. Input/output data management
    3. Processing step recording
    4. Intent tracking
    5. Type-safe resource management

    Attributes:
        info (Dict[str, Any]): General information about the resource.
        intent (str): The intent associated with the resource.
        input (List[Any]): Input data for the resource.
        output (str): Output data from the resource.
        steps (List[Any]): Processing steps taken.
        personalized_intent (str): User-specific intent for the resource.
    """

    info: dict[str, Any]
    intent: str = Field(default="")
    input: list[Any] = Field(default_factory=list)
    output: str = Field(default="")
    steps: list[Any] = Field(default_factory=list)
    personalized_intent: str = Field(default="")


class Memory(BaseModel):
    """Memory management for conversation and processing.

    This class maintains the memory of the conversation system, including
    processing trajectories and function call history.

    The class provides:
    1. Trajectory tracking
    2. Function call history
    3. Type-safe memory management

    Attributes:
        trajectory (List[List[ResourceRecord]]): Processing trajectory history.
        function_calling_trajectory (List[Dict[str, Any]]): Function call history.
    """

    trajectory: list[list[ResourceRecord]] = Field(default_factory=list)
    function_calling_trajectory: list[dict[str, Any]] = Field(default_factory=list)



from typing import Any

from pydantic import BaseModel

from arklex.env.workers.base.entities import WorkerOutput
from arklex.orchestrator.entities.orchestrator_state_entities import (
    StatusEnum,
)


class MilvusRAGWorkerData(BaseModel):
    """Data for the Milvus RAG worker."""

    tags: dict[str, Any]


class MilvusRAGWorkerOutput(WorkerOutput):
    """Response for the Milvus RAG worker."""

    response: str
    status: StatusEnum



from agents import Agent, Tool

from arklex.env.agents.utils.tool_resolver import resolve_tools_for_agent
from arklex.utils.llm_config import LLMConfig


def build_agents(agent_configs: list[dict], llm_config: LLMConfig) -> list[Agent]:
    agents = []
    for config in agent_configs:
        tools = resolve_tools_for_agent(config.get("tools", []))
        agent = Agent(
            name=config["name"],
            instructions=config["task"],
            tools=tools,
            model=llm_config.model_type_or_path,
        )
        agents.append(agent)
    return agents


def build_tool_wrapped_agents(
    agent_configs: list[dict], llm_config: LLMConfig
) -> tuple[list[Agent], list[Tool]]:
    """
    Builds agents using `build_agents`, then wraps each as a tool.

    Returns:
        (agents, tool_wrappers)
    """
    agents = build_agents(agent_configs, llm_config)
    tool_wrappers = [
        agent.as_tool(
            tool_name=agent.name,
            tool_description=cfg.get("description", f"Tool for {cfg['name']}"),
        )
        for agent, cfg in zip(agents, agent_configs, strict=False)
    ]
    return agents, tool_wrappers



from datetime import datetime
from enum import Enum


class Transcript:
    def __init__(self, id: str, text: str, origin: str, created_at: datetime) -> None:
        self.id = id
        self.text = text
        self.origin = origin
        self.created_at = created_at


# Enum for chat roles
class ChatRole(str, Enum):
    BOT = "bot"
    ASSISTANT = "assistant"
    USER = "user"
    BOT_FOLLOWUP = "bot_followup"
    HUMAN_AGENT = "human_agent"
    TOOL = "tool"

    def from_str(role: str) -> "ChatRole":
        if role == "bot":
            return ChatRole.BOT
        elif role == "assistant":
            return ChatRole.ASSISTANT
        elif role == "user":
            return ChatRole.USER
        elif role == "bot_followup":
            return ChatRole.BOT_FOLLOWUP
        elif role == "human_agent":
            return ChatRole.HUMAN_AGENT
        elif role == "tool":
            return ChatRole.TOOL
        else:
            raise ValueError(f"Invalid role: {role}")


class ResourceAuthGroup(int, Enum):
    """when adding new auth group, add it also in the backend sql script"""

    PUBLIC = -1
    GOOGLE_CALENDAR = 0
    SHOPIFY = 1
    HUBSPOT = 2
    TWILIO = 3
    SALESFORCE = 4



"""UI components for the Arklex task graph generator.

This module contains interactive UI components for editing and managing tasks.
"""

from . import task_editor
from .data_manager import TaskDataManager
from .input_modal import InputModal
from .protocols import (
    InputModalProtocol,
    TreeNodeProtocol,
    TreeProtocol,
)
from .task_editor import TaskEditorApp

__all__ = [
    "TaskEditorApp",
    "InputModal",
    "task_editor",
    "TreeNodeProtocol",
    "TreeProtocol",
    "InputModalProtocol",
    "TaskDataManager",
]



"""Script for analyzing TAU benchmark results.

This script provides functionality to analyze the results of TAU benchmark runs,
including error identification and result processing. It sets up the necessary
environment and runs the error identification process with default parameters
for the retail environment and OpenAI platform.
"""

import argparse
import os
import sys

from dotenv import load_dotenv

from benchmark.tau_bench.auto_error_identification import (
    get_args,
    run_error_identification,
)

root_dir: str = os.path.abspath(os.path.join(os.path.dirname(__file__), "..", ".."))
sys.path.append(root_dir)
load_dotenv()

if __name__ == "__main__":
    """Main entry point for the TAU benchmark result analysis.
    
    This script runs the error identification process with the following default parameters:
    - Environment: retail
    - Platform: OpenAI
    - Max concurrency: 16
    
    Additional parameters can be provided via command line arguments:
    --results-path: Path to the results directory
    --output-dir: Path to the output directory
    """
    sys.argv += ["--env", "retail"]
    sys.argv += ["--platform", "openai"]
    sys.argv += ["--max-concurrency", "16"]
    args: argparse.Namespace = get_args()
    run_error_identification(args)



"""
This module provides functionality to add items to a shopping cart in the Shopify store.
It supports adding multiple product variants to a cart with specified quantities.

Module Name: cart_add_items

This file contains the code for adding items to a shopping cart.
"""

import inspect
import json

import requests
from pydantic import BaseModel

from arklex.env.tools.shopify._exception_prompt import ShopifyExceptionPrompt
from arklex.env.tools.shopify.base.entities import ShopifyStorefrontAuth
from arklex.env.tools.shopify.utils.utils import authorify_storefront
from arklex.env.tools.shopify.utils.utils_slots import (
    ShopifyCartAddItemsSlots,
)
from arklex.env.tools.tools import register_tool
from arklex.utils.exceptions import ToolExecutionError
from arklex.utils.logging_utils import LogContext

log_context = LogContext(__name__)


class CartAddItemsOutput(BaseModel):
    message_flow: str


@register_tool(
    description="Add items to user's shopping cart.",
    slots=ShopifyCartAddItemsSlots.get_all_slots(),
)
def cart_add_items(
    cart_id: str,
    product_variant_ids: list[str],
    auth: ShopifyStorefrontAuth,
    **kwargs: object,
) -> CartAddItemsOutput:
    """
    Add items to a shopping cart.

    Args:
        cart_id (str): The ID of the shopping cart.
        product_variant_ids (list[str]): List of product variant IDs to add to the cart.
        auth (ShopifyStorefrontAuth): Authentication credentials for the Shopify store.
        **kwargs: Additional keyword arguments for llm configuration.

    Returns:
        CartAddItemsOutput: A success message with the cart details if successful.

    Raises:
        ToolExecutionError: If:
            - The items cannot be added to the cart
            - There are errors in the cart data
            - There's an error in the request process
    """
    func_name = inspect.currentframe().f_code.co_name
    auth = authorify_storefront(auth)

    variable: dict[str, list[dict[str, str | int]]] = {
        "cartId": cart_id,
        "lines": [
            {"merchandiseId": pv_id, "quantity": 1} for pv_id in product_variant_ids
        ],
    }
    headers: dict[str, str] = {
        "X-Shopify-Storefront-Access-Token": auth["storefront_token"]
    }
    query = """
    mutation cartLinesAdd($cartId: ID!, $lines: [CartLineInput!]!) {
        cartLinesAdd(cartId: $cartId, lines: $lines) {
            cart {
                checkoutUrl
            }
        }
    }
    """
    try:
        response = requests.post(
            auth["storefront_url"],
            json={"query": query, "variables": variable},
            headers=headers,
        )
        if response.status_code == 200:
            cart_data = response.json()
            if "errors" in cart_data:
                raise ToolExecutionError(
                    func_name,
                    extra_message=ShopifyExceptionPrompt.CART_ADD_ITEMS_ERROR_PROMPT,
                )
            else:
                # Check for required keys in the response
                if "data" not in cart_data:
                    raise ToolExecutionError(
                        func_name,
                        extra_message=ShopifyExceptionPrompt.CART_ADD_ITEMS_ERROR_PROMPT,
                    )

                if "cartLinesAdd" not in cart_data["data"]:
                    raise ToolExecutionError(
                        func_name,
                        extra_message=ShopifyExceptionPrompt.CART_ADD_ITEMS_ERROR_PROMPT,
                    )

                cart_lines_add = cart_data["data"]["cartLinesAdd"]
                if "cart" not in cart_lines_add or cart_lines_add["cart"] is None:
                    raise ToolExecutionError(
                        func_name,
                        extra_message=ShopifyExceptionPrompt.CART_ADD_ITEMS_ERROR_PROMPT,
                    )

                return CartAddItemsOutput(
                    message_flow="Items are successfully added to the shopping cart. "
                    + json.dumps(cart_lines_add["cart"]),
                )
        else:
            raise ToolExecutionError(
                func_name,
                extra_message=ShopifyExceptionPrompt.CART_ADD_ITEMS_ERROR_PROMPT,
            )
    except (requests.RequestException, Exception) as e:
        log_context.error(f"Cart add items failed: {e}")
        raise ToolExecutionError(
            func_name,
            extra_message=ShopifyExceptionPrompt.CART_ADD_ITEMS_ERROR_PROMPT,
        ) from e



"""Base worker implementation for the Arklex framework.

This module provides the base abstract class for workers in the framework. It defines the
interface for worker execution and includes functionality for worker registration, error
handling, and state management. The module serves as a foundation for implementing specific
workers that handle different types of tasks and operations within the system.
"""

from abc import ABC, abstractmethod
from typing import Any

from arklex.env.workers.base.entities import WorkerOutput
from arklex.orchestrator.entities.orchestrator_state_entities import (
    OrchestratorState,
)
from arklex.utils.logging_utils import LogContext

log_context = LogContext(__name__)


class BaseWorker(ABC):
    """Base abstract class for workers in the Arklex framework.

    This class defines the interface for worker execution and provides common
    functionality for all workers. It includes methods for string representation,
    execution handling, and error management.

    Attributes:
        description (Optional[str]): Description of the worker's functionality.
    """

    description: str | None = None

    def __str__(self) -> str:
        """Get a string representation of the worker.

        Returns:
            str: The name of the worker class.
        """
        return f"{self.__class__.__name__}"

    def __repr__(self) -> str:
        """Get a detailed string representation of the worker.

        Returns:
            str: The name of the worker class.
        """
        return f"{self.__class__.__name__}"

    @abstractmethod
    def init_worker_data(
        self, orch_state: OrchestratorState, node_specific_data: dict[str, Any]
    ) -> None:
        """Initialize the worker data.

        Args:
            orch_state (OrchestratorState): The current orchestrator state.
            node_specific_data (dict[str, Any]): Additional keyword arguments for the execution.
        """

    @abstractmethod
    def _execute(self) -> WorkerOutput:
        """Execute the worker's core functionality.

        This abstract method must be implemented by concrete worker classes to
        define their specific execution logic.

        Returns:
            WorkerOutput: The execution results.
        """

    def execute(
        self,
        orch_state: OrchestratorState,
        node_specific_data: dict[str, Any],
    ) -> tuple[OrchestratorState, WorkerOutput]:
        """Execute the worker with error handling and state management.

        This method wraps the worker's execution with error handling and state
        management. It processes the execution results and updates the message state.

        Args:
            orch_state (OrchestratorState): The current orchestrator state.
            node_specific_data (dict[str, Any]): Additional keyword arguments for the execution.

        Returns:
            MessageState: The updated message state after execution.
        """
        self.init_worker_data(orch_state, node_specific_data)
        worker_output: WorkerOutput = self._execute()

        return self.orch_state, worker_output



# TODO(christian): put annotations correctly.

import abc
import enum
import json
from typing import Any, TypeVar

from pydantic import BaseModel

from benchmark.tau_bench.model_utils.api.datapoint import (
    BinaryClassifyDatapoint,
    ClassifyDatapoint,
    Datapoint,
    GenerateDatapoint,
    ParseDatapoint,
    ParseForceDatapoint,
    ScoreDatapoint,
)
from benchmark.tau_bench.model_utils.api.types import PartialObj
from benchmark.tau_bench.model_utils.model.exception import ModelError
from benchmark.tau_bench.model_utils.model.general_model import GeneralModel
from benchmark.tau_bench.model_utils.model.utils import (
    add_md_tag,
    clean_top_level_keys,
    display_choices,
    json_response_to_obj_or_partial_obj,
    optionalize_type,
    parse_json_or_json_markdown,
    try_classify_recover,
    type_to_json_schema_string,
)

T = TypeVar("T", bound=BaseModel)


class Role(str, enum.Enum):
    SYSTEM = "system"
    ASSISTANT = "assistant"
    USER = "user"


class Message(BaseModel):
    role: Role
    content: str
    obj: dict[str, Any] | None = None

    def model_dump(self, **kwargs: object) -> dict[str, Any]:
        if self.obj is not None:
            return super().model_dump(**kwargs)
        return {"role": self.role, "content": self.content}


class PromptSuffixStrategy(str, enum.Enum):
    JSON = "json"
    JSON_MD_BLOCK = "json_md_block"


def force_json_prompt(
    text: str,
    suffix_strategy: PromptSuffixStrategy = PromptSuffixStrategy.JSON,
) -> str:
    if suffix_strategy == PromptSuffixStrategy.JSON:
        return f"{text}\n\nValid JSON:"
    elif suffix_strategy == PromptSuffixStrategy.JSON_MD_BLOCK:
        return f'{text}\n\nThe result should be a valid JSON object (according to the definition in the provided schema) in a markdown block only. For example:\nassistant:```json\n{{"items": ["value"]}}\n```'
    else:
        raise ValueError(f"Invalid suffix strategy: {suffix_strategy}")


def build_generate_state(
    instruction: str,
    text: str,
    examples: list[GenerateDatapoint] | None = None,
) -> list[Message]:
    messages = []
    if examples is not None:
        for example in examples:
            example_msgs = [
                Message(role=Role.SYSTEM, content=example.instruction),
                Message(role=Role.USER, content=example.text),
                Message(role=Role.ASSISTANT, content=example.response),
            ]
            messages.extend(example_msgs)
    messages.append(Message(role=Role.SYSTEM, content=instruction))
    messages.append(Message(role=Role.USER, content=text))
    return messages


def build_parse_force_state(
    instruction: str,
    typ: type[T] | dict[str, Any],
    text: str | None = None,
    examples: list[ParseForceDatapoint] | None = None,
    suffix_strategy: PromptSuffixStrategy = PromptSuffixStrategy.JSON,
) -> list[Message]:
    def display_sample(
        instr: str,
        ty: type[T] | dict[str, Any],
        t: str | None = None,
        response: T | dict[str, Any] | None = None,
    ) -> Message | list[Message]:
        if isinstance(ty, dict):
            json_schema_string = json.dumps(ty)
        else:
            json_schema_string = type_to_json_schema_string(ty)
        text_insert = "" if t is None else f"\n\nText:\n{t}"
        input_text = force_json_prompt(
            text=f"Instruction:\n{instr}{text_insert}\n\nSchema:\n{json_schema_string}",
            suffix_strategy=suffix_strategy,
        )
        if response is not None:
            if isinstance(response, dict):
                response_display = json.dumps(response)
            else:
                response_display = json.dumps(response.model_dump())
            return [
                Message(role=Role.USER, content=input_text),
                Message(role=Role.ASSISTANT, content=response_display),
            ]
        else:
            return Message(role=Role.USER, content=input_text)

    messages = [
        Message(
            role=Role.SYSTEM,
            content="Generate an object with the provided instruction, text, and schema.",
        ),
    ]
    if examples is not None:
        for example in examples:
            example_msgs = display_sample(
                instr=example.instruction,
                ty=example.typ,
                t=example.text,
                response=example.response,
            )
            assert isinstance(example_msgs, list) and all(
                isinstance(msg, Message) for msg in example_msgs
            )
            messages.extend(example_msgs)
    messages.append(display_sample(instr=instruction, ty=typ, t=text))
    return messages


def build_score_state(
    instruction: str,
    text: str,
    min: int,
    max: int,
    examples: list[ScoreDatapoint] | None = None,
    suffix_strategy: PromptSuffixStrategy = PromptSuffixStrategy.JSON,
) -> list[Message]:
    def display_sample(
        instr: str, t: str, mn: int, mx: int, response: int | None = None
    ) -> list[Message] | Message:
        if mn > mx:
            raise ValueError(f"Invalid range: [{mn}, {mx}]")
        input_text = force_json_prompt(
            f"Instruction:\n{instr}\n\nText:\n{t}\n\nRange:\n[{mn}, {mx}]",
            suffix_strategy,
        )
        if response is not None:
            return [
                Message(role=Role.USER, content=input_text),
                Message(role=Role.ASSISTANT, content=f'{{"score": {response}}}'),
            ]
        else:
            return Message(role=Role.USER, content=input_text)

    messages = [
        Message(
            role=Role.SYSTEM,
            content='Score the following text with the provided instruction and range as an integer value in valid JSON:\n{"score": number}',
        ),
    ]
    if examples is not None:
        for example in examples:
            example_msgs = display_sample(
                instr=example.instruction,
                t=example.text,
                mn=example.min,
                mx=example.max,
                response=example.response,
            )
            assert isinstance(example_msgs, list) and all(
                isinstance(msg, Message) for msg in example_msgs
            ), example_msgs
            messages.extend(example_msgs)
    messages.append(display_sample(instr=instruction, t=text, mn=min, mx=max))
    return messages


def build_parse_state(
    text: str,
    typ: type[T] | dict[str, Any],
    examples: list[ParseDatapoint] | None = None,
    suffix_strategy: PromptSuffixStrategy = PromptSuffixStrategy.JSON,
) -> list[Message]:
    def display_sample(
        t: str,
        ty: type[T] | dict[str, Any],
        response: T | PartialObj | dict[str, Any] | None = None,
    ) -> Message | list[Message]:
        if isinstance(ty, dict):
            json_schema_string = json.dumps(ty)
        else:
            optionalized_typ = optionalize_type(ty)
            json_schema_string = type_to_json_schema_string(optionalized_typ)
        input_text = force_json_prompt(
            f"Text:\n{t}\n\nSchema:\n{json_schema_string}",
            suffix_strategy=suffix_strategy,
        )
        if response is not None:
            if isinstance(response, dict):
                response_display = json.dumps(response)
            else:
                response_display = response.model_dump_json()
            return [
                Message(role=Role.USER, content=input_text),
                Message(role=Role.ASSISTANT, content=response_display),
            ]
        else:
            return Message(role=Role.USER, content=input_text)

    messages = [
        Message(
            role=Role.SYSTEM,
            content="Parse the following text with the provided JSON schema.",
        ),
    ]
    if examples is not None:
        for example in examples:
            example_msgs = display_sample(
                t=example.text, ty=typ, response=example.response
            )
            assert isinstance(example_msgs, list) and all(
                isinstance(msg, Message) for msg in example_msgs
            ), example_msgs
            messages.extend(example_msgs)
    messages.append(display_sample(t=text, ty=typ))
    return messages


def build_classify_state(
    instruction: str,
    text: str,
    options: list[str],
    examples: list[ClassifyDatapoint] | None = None,
    suffix_strategy: PromptSuffixStrategy = PromptSuffixStrategy.JSON,
) -> tuple[list[Message], dict[str, int]]:
    def display_sample(
        instr: str, t: str, opts: list[str], response: int | None = None
    ) -> list[Message] | tuple[Message, dict[str, int]]:
        choices_display, decode_map = display_choices(opts)
        input_text = force_json_prompt(
            f"Instruction:\n{instr}\n\nText:\n{t}\n\nChoices:\n{choices_display}",
            suffix_strategy=suffix_strategy,
        )
        if response is not None:
            response_label = None
            for label, idx in decode_map.items():
                if idx == response:
                    response_label = label
                    break
            assert response_label is not None, f"Invalid response: {response}"
            return [
                Message(role=Role.USER, content=input_text),
                Message(
                    role=Role.ASSISTANT,
                    content=f'{{"classification": "{response_label}"}}',
                ),
            ]
        else:
            return Message(role=Role.USER, content=input_text), decode_map

    messages = [
        Message(
            role=Role.SYSTEM,
            content='Classify the following text with the provided instruction and choices. To classify, provide the key of the choice:\n{"classification": string}\n\nFor example, if the correct choice is \'Z. description of choice Z\', then provide \'Z\' as the classification as valid JSON:\n{"classification": "Z"}',
        ),
    ]
    if examples is not None:
        for example in examples:
            example_msgs = display_sample(
                instr=example.instruction,
                t=example.text,
                opts=example.options,
                response=example.response,
            )
            assert isinstance(example_msgs, list) and all(
                isinstance(msg, Message) for msg in example_msgs
            ), example_msgs
            messages.extend(example_msgs)
    message, decode_map = display_sample(instr=instruction, t=text, opts=options)
    messages.append(message)
    return messages, decode_map


class ChatModel(GeneralModel):
    @abc.abstractmethod
    def generate_message(
        self,
        messages: list[Message],
        force_json: bool,
        temperature: float | None = None,
    ) -> Message:
        raise NotImplementedError

    def handle_generate_message_response(
        self, prompt: list[dict[str, str] | Message], content: str, force_json: bool
    ) -> Message:
        if force_json:
            try:
                parsed = parse_json_or_json_markdown(content)
            except (json.JSONDecodeError, ValueError) as e:
                msgs = []
                for msg in prompt:
                    if isinstance(msg, Message):
                        msgs.append(msg.model_dump())
                    else:
                        msgs.append(msg)
                raise ModelError(
                    short_message=f"Failed to parse JSON: {content}",
                    prompt=msgs,
                    response=content,
                ) from e
            cleaned = clean_top_level_keys(parsed)
            return Message(role=Role.ASSISTANT, content=content, obj=cleaned)
        return Message(role=Role.ASSISTANT, content=content, obj=None)

    def build_generate_message_state(
        self, messages: list[Message]
    ) -> list[dict[str, str]]:
        msgs: list[dict[str, str]] = []
        for msg in messages:
            content = json.dumps(msg.obj) if msg.obj is not None else msg.content
            msgs.append({"role": msg.role.value, "content": content})
        return msgs

    def _handle_classify_response(
        self, res: Message, decode_map: dict[str, int]
    ) -> int:
        assert res.obj is not None
        if "classification" not in res.obj:
            raise ModelError(f"Invalid response from model: {res.content}")
        choice = res.obj["classification"]
        if choice not in decode_map:
            key = try_classify_recover(s=choice, decode_map=decode_map)
            if key is not None:
                return decode_map[key]
            raise ModelError(f"Invalid choice: {choice}")
        return decode_map[choice]

    def classify(
        self,
        instruction: str,
        text: str,
        options: list[str],
        examples: list[ClassifyDatapoint] | None = None,
        temperature: float | None = None,
    ) -> int:
        messages, decode_map = build_classify_state(
            instruction, text, options, examples=examples
        )
        res = self.generate_message(messages, force_json=True, temperature=temperature)
        return self._handle_classify_response(res, decode_map)

    def parse(
        self,
        text: str,
        typ: type[T] | dict[str, Any],
        examples: list[ParseDatapoint] | None = None,
        temperature: float | None = None,
    ) -> T | PartialObj | dict[str, Any]:
        messages = build_parse_state(text, typ, examples=examples)
        res = self.generate_message(messages, force_json=True, temperature=temperature)
        assert res.obj is not None
        return json_response_to_obj_or_partial_obj(response=res.obj, typ=typ)

    def generate(
        self,
        instruction: str,
        text: str,
        examples: list[GenerateDatapoint] | None = None,
        temperature: float | None = None,
    ) -> str:
        messages = build_generate_state(
            instruction=instruction, text=text, examples=examples
        )
        return self.generate_message(
            messages, force_json=False, temperature=temperature
        ).content

    def _handle_parse_force_response(
        self, res: Message, typ: type[T] | dict[str, Any]
    ) -> T | dict[str, Any]:
        assert res.obj is not None
        obj = json_response_to_obj_or_partial_obj(response=res.obj, typ=typ)
        if not isinstance(typ, dict) and isinstance(obj, dict):
            raise ModelError(f"Invalid response from model: {res.content}")
        return obj

    def parse_force(
        self,
        instruction: str,
        typ: type[T] | dict[str, Any],
        text: str | None = None,
        examples: list[ParseForceDatapoint] | None = None,
        temperature: float | None = None,
    ) -> T | dict[str, Any]:
        messages = build_parse_force_state(
            instruction=instruction,
            typ=typ,
            text=text,
            examples=examples,
        )
        res = self.generate_message(messages, force_json=True, temperature=temperature)
        return self._handle_parse_force_response(res, typ)

    def _handle_score_response(
        self,
        res: Message,
        min: int,
        max: int,
    ) -> int:
        if res.obj is None or "score" not in res.obj:
            raise ModelError(f"Invalid response from model: {res.content}")
        score = res.obj["score"]
        if not isinstance(score, int):
            raise ModelError(f"Invalid score type: {type(score)}")
        if score < min or score > max:
            raise ModelError(f"Invalid score value: {score}")
        return score

    def score(
        self,
        instruction: str,
        text: str,
        min: int,
        max: int,
        examples: list[ScoreDatapoint] | None = None,
        temperature: float | None = None,
    ) -> int:
        messages = build_score_state(instruction, text, min, max, examples=examples)
        res = self.generate_message(messages, force_json=True, temperature=temperature)
        return self._handle_score_response(res, min, max)


def build_prompts(
    dps: list[Datapoint], prompt_suffix_strategy: PromptSuffixStrategy | None
) -> list[str | list[Message]]:
    if len(dps) == 0:
        return []
    typ = type(dps[0])
    for i, dp in enumerate(dps):
        if not isinstance(dp, typ):
            raise ValueError(
                f"All elements must be of type Datapoint, expected type {typ} at index {i}, got {type(dp)}"
            )
    if isinstance(dps[0], ParseDatapoint):
        build_func = build_parse_prompts
    elif isinstance(dps[0], BinaryClassifyDatapoint):
        build_func = build_binary_classify_prompts
    elif isinstance(dps[0], ClassifyDatapoint):
        build_func = build_classify_prompts
    elif isinstance(dps[0], ParseForceDatapoint):
        build_func = build_parse_force_prompts
    elif isinstance(dps[0], GenerateDatapoint):
        build_func = build_generate_prompts
    elif isinstance(dps[0], ScoreDatapoint):
        build_func = build_score_prompts
    else:
        raise ValueError(f"Unknown datapoint type: {type(dps[0])}")
    return build_func(dps, suffix_strategy=prompt_suffix_strategy)


def build_parse_prompts(
    dps: list[ParseDatapoint],
    suffix_strategy: PromptSuffixStrategy | None = None,
) -> list[str | list[Message]]:
    datapoints = []
    for dp in dps:
        json_response_object = (
            dp.response.model_dump_json()
            if isinstance(dp.response, BaseModel)
            else json.dumps(dp.response)
        )
        prompt_msgs = build_parse_state(
            text=dp.text,
            typ=dp.typ,
            suffix_strategy=(
                suffix_strategy
                if suffix_strategy is not None
                else PromptSuffixStrategy.JSON
            ),
        )
        json_response = apply_suffix_strategy(
            response=json_response_object, suffix_strategy=suffix_strategy
        )
        datapoints.append(
            prompt_msgs + [Message(role=Role.ASSISTANT, content=json_response)]
        )
    return datapoints


def build_binary_classify_prompts(
    dps: list[BinaryClassifyDatapoint],
    suffix_strategy: PromptSuffixStrategy | None = None,
) -> list[str | list[Message]]:
    return build_classify_prompts(
        [
            ClassifyDatapoint(
                instruction=dp.instruction,
                text=dp.text,
                options=["true", "false"],
                response=0 if dp.response else 1,
            )
            for dp in dps
        ],
        suffix_strategy=suffix_strategy,
    )


def build_classify_prompts(
    dps: list[ClassifyDatapoint],
    suffix_strategy: PromptSuffixStrategy | None = None,
) -> list[str | list[Message]]:
    def label_idx_to_label_json(idx: int, decode_map: dict[str, int]) -> str:
        label = None
        for k, v in decode_map.items():
            if v == idx:
                label = k
                break
        if label is None:
            raise ValueError(f"Label index {idx} not found in decode map")
        return f'{{"classification": "{label}"}}'

    datapoints = []
    for dp in dps:
        suffix_strategy = (
            PromptSuffixStrategy.JSON if suffix_strategy is None else suffix_strategy
        )
        prompt_msgs, decode_map = build_classify_state(
            instruction=dp.instruction,
            text=dp.text,
            options=dp.options,
            suffix_strategy=suffix_strategy,
        )
        json_response_object = label_idx_to_label_json(
            idx=dp.response, decode_map=decode_map
        )
        json_response = apply_suffix_strategy(
            response=json_response_object, suffix_strategy=suffix_strategy
        )
        datapoints.append(
            prompt_msgs
            + [
                Message(
                    role=Role.ASSISTANT,
                    content=json_response,
                )
            ]
        )
    return datapoints


def build_parse_force_prompts(
    dps: list[ParseForceDatapoint],
    suffix_strategy: PromptSuffixStrategy | None = None,
) -> list[str | list[Message]]:
    datapoints = []
    for dp in dps:
        json_response_obj = (
            dp.response.model_dump_json()
            if isinstance(dp.response, BaseModel)
            else json.dumps(dp.response)
        )
        suffix_strategy = (
            PromptSuffixStrategy.JSON if suffix_strategy is None else suffix_strategy
        )
        prompt_msgs = build_parse_force_state(
            instruction=dp.instruction,
            text=dp.text,
            typ=dp.typ,
            suffix_strategy=suffix_strategy,
        )
        json_response = apply_suffix_strategy(
            response=json_response_obj, suffix_strategy=suffix_strategy
        )
        datapoints.append(
            prompt_msgs + [Message(role=Role.ASSISTANT, content=json_response)]
        )
    return datapoints


def build_generate_prompts(dps: list[GenerateDatapoint]) -> list[str | list[Message]]:
    datapoints = []
    for dp in dps:
        prompt_msgs = build_generate_state(instruction=dp.instruction, text=dp.text)
        datapoints.append(
            prompt_msgs + [Message(role=Role.ASSISTANT, content=dp.response)]
        )
    return datapoints


def build_score_prompts(
    dps: list[ScoreDatapoint],
    suffix_strategy: PromptSuffixStrategy | None = None,
) -> list[str | list[Message]]:
    datapoints = []
    for dp in dps:
        json_response_object = f'{{"score": {dp.response}}}'
        suffix_strategy = (
            suffix_strategy
            if suffix_strategy is not None
            else PromptSuffixStrategy.JSON
        )
        prompt_msgs = build_score_state(
            instruction=dp.instruction,
            text=dp.text,
            min=dp.min,
            max=dp.max,
            suffix_strategy=suffix_strategy,
        )
        json_response = apply_suffix_strategy(
            response=json_response_object, suffix_strategy=suffix_strategy
        )
        datapoints.append(
            prompt_msgs + [Message(role=Role.ASSISTANT, content=json_response)]
        )
    return datapoints


def apply_suffix_strategy(response: str, suffix_strategy: PromptSuffixStrategy) -> str:
    if suffix_strategy == PromptSuffixStrategy.JSON:
        return response
    elif suffix_strategy == PromptSuffixStrategy.JSON_MD_BLOCK:
        return add_md_tag(response)
    else:
        raise ValueError(f"Unknown suffix strategy: {suffix_strategy}")



"""Redis connection and pool management for the Arklex framework.

This module provides functionality for managing Redis connections using connection pooling.
It includes configuration settings for Redis connections, connection pool management,
and utility methods for executing Redis commands and managing cache operations.
The module implements connection pooling to efficiently handle multiple Redis operations
while maintaining connection limits and proper resource management.

The module is organized into several key components:
1. Configuration: Environment-based Redis configuration settings
2. Connection Pool: Management of Redis connection pools
3. Command Execution: Methods for executing Redis commands
4. Cache Operations: Utility methods for common caching patterns
5. Resource Management: Proper handling of Redis connections and resources

Key Features:
- Connection pooling for efficient resource management
- Environment-based configuration
- Automatic connection timeout handling
- Comprehensive error handling and logging
- Support for various Redis data types
- Cache utility methods with TTL support
- Resource cleanup

Usage:
    from arklex.utils.redis import RedisPool, redis_pool

    # Use the global instance
    redis_pool.set("key", "value", ttl=3600)
    value = redis_pool.get("key")

    # Or create a custom instance
    custom_pool = RedisPool(
        host="localhost",
        port=6379,
        db=0,
        password="password",
        max_connections=10
    )
"""

import json
import os
from typing import Any

import redis
from redis.connection import ConnectionPool

from arklex.utils.logging_utils import LogContext

log_context = LogContext(__name__)

# Timeout in seconds for establishing Redis connections
CONNECTION_TIMEOUT: int = int(os.getenv("REDIS_CONNECTION_TIMEOUT", 5))

# Redis configuration loaded from environment variables
REDIS_CONFIG: dict[str, str | int | None] = {
    "host": os.getenv("REDIS_HOST", "localhost"),  # Redis host address
    "port": int(os.getenv("REDIS_PORT", 6379)),  # Redis port number
    "db": int(os.getenv("REDIS_DB", 0)),  # Redis database number
    "password": os.getenv("REDIS_PASSWORD"),  # Redis password (optional)
    "username": os.getenv("REDIS_USERNAME"),  # Redis username (optional)
}

# Maximum number of connections in the connection pool
POOL_SIZE: int = int(os.getenv("REDIS_POOL_SIZE", 10))

# Default TTL for cache operations (in seconds)
DEFAULT_TTL: int = int(os.getenv("REDIS_DEFAULT_TTL", 3600))


class RedisPool:
    """A connection pool manager for Redis connections.

    This class manages a pool of Redis connections, providing methods for
    obtaining connections, executing commands, and managing cache operations. It implements
    connection pooling to efficiently handle multiple Redis operations while
    maintaining connection limits and proper resource management.

    The class provides the following key features:
    1. Connection Pool Management: Efficient handling of Redis connections
    2. Command Execution: Methods for executing Redis commands with proper resource cleanup
    3. Cache Operations: High-level methods for common caching patterns
    4. Error Handling: Comprehensive error handling and logging
    5. Resource Cleanup: Automatic cleanup of Redis resources

    Attributes:
        _host (str): The Redis host address
        _port (int): The Redis port number
        _db (int): The Redis database number
        _password (Optional[str]): The Redis password
        _username (Optional[str]): The Redis username
        redis_config (Dict[str, Any]): Complete Redis configuration dictionary
        connection_pool (ConnectionPool): The Redis connection pool instance
        client (redis.Redis): The Redis client instance
    """

    def __init__(
        self, max_connections: int = POOL_SIZE, **kwargs: dict[str, Any]
    ) -> None:
        """Initialize the Redis connection pool.

        This method sets up a connection pool with the specified size and configuration.
        It initializes the Redis connection parameters and creates the connection pool.

        Args:
            max_connections (int): The maximum number of connections in the pool
            **kwargs (Dict[str, Any]): Redis configuration parameters including:
                - host: Redis host address (default: "localhost")
                - port: Redis port number (default: 6379)
                - db: Redis database number (default: 0)
                - password: Redis password (default: None)
                - username: Redis username (default: None)
        """
        self._host: str = kwargs.get("host", "localhost")
        self._port: int = kwargs.get("port", 6379)
        self._db: int = kwargs.get("db", 0)
        self._password: str | None = kwargs.get("password")
        self._username: str | None = kwargs.get("username")

        # Build Redis configuration
        self.redis_config: dict[str, Any] = {
            "host": self._host,
            "port": self._port,
            "db": self._db,
            "socket_timeout": CONNECTION_TIMEOUT,
            "socket_connect_timeout": CONNECTION_TIMEOUT,
            "decode_responses": False,  # Keep as bytes for flexibility
        }

        # Add optional authentication parameters
        if self._password:
            self.redis_config["password"] = self._password
        if self._username:
            self.redis_config["username"] = self._username

        # Create connection pool
        self.connection_pool = ConnectionPool(
            max_connections=max_connections, **self.redis_config
        )

        # Create Redis client
        self.client = redis.Redis(connection_pool=self.connection_pool)

        log_context.info(
            f"Redis pool initialized with {max_connections} max connections"
        )

    def get_connection(self) -> redis.Redis:
        """Get a Redis client instance.

        This method returns the Redis client which manages connections automatically
        through the connection pool.

        Returns:
            redis.Redis: A Redis client instance
        """
        return self.client

    def ping(self) -> bool:
        """Test the Redis connection.

        This method tests the connection to Redis by sending a PING command.

        Returns:
            bool: True if the connection is successful, False otherwise
        """
        try:
            response = self.client.ping()
            log_context.debug("Redis ping successful")
            return response
        except Exception as e:
            log_context.error(f"Redis ping failed: {e}")
            return False

    def set(
        self,
        key: str,
        value: str | bytes | dict | list | int | float | bool,
        ttl: int | None = None,
    ) -> bool:
        """Set a key-value pair in Redis with optional TTL.

        This method stores a value in Redis with the given key. If the value is not
        a string or bytes, it will be JSON-serialized.

        Args:
            key (str): The Redis key
            value (str | bytes | dict | list | int | float | bool): The value to store
            ttl (Optional[int]): Time to live in seconds (default: None for no expiration)

        Returns:
            bool: True if the operation was successful, False otherwise
        """
        try:
            # Serialize non-string/bytes values as JSON
            if isinstance(value, str | bytes):
                stored_value = value
            else:
                stored_value = json.dumps(value)

            if ttl:
                result = self.client.setex(key, ttl, stored_value)
            else:
                result = self.client.set(key, stored_value)

            log_context.debug(f"Redis SET successful for key: {key}")
            return bool(result)
        except Exception as e:
            log_context.error(f"Redis SET failed for key {key}: {e}")
            return False

    def get(self, key: str, decode_json: bool = True) -> str | None:
        """Get a value from Redis by key.

        This method retrieves a value from Redis. If decode_json is True and the value
        appears to be JSON, it will be deserialized.

        Args:
            key (str): The Redis key
            decode_json (bool): Whether to attempt JSON deserialization (default: True)

        Returns:
            str | None: The retrieved value as a string, or None if the key doesn't exist
        """
        try:
            value = self.client.get(key)
            if value is None:
                return None

            # Decode bytes to string
            if isinstance(value, bytes):
                value = value.decode("utf-8")

            # Attempt JSON deserialization if requested
            if decode_json and isinstance(value, str):
                try:
                    return json.loads(value)
                except (json.JSONDecodeError, ValueError):
                    # Return as string if not valid JSON
                    pass

            log_context.debug(f"Redis GET successful for key: {key}")
            return value
        except Exception as e:
            log_context.error(f"Redis GET failed for key {key}: {e}")
            return None

    def delete(self, *keys: str) -> int:
        """Delete one or more keys from Redis.

        Args:
            *keys (str): One or more Redis keys to delete

        Returns:
            int: Number of keys that were deleted
        """
        try:
            result = self.client.delete(*keys)
            log_context.debug(f"Redis DELETE successful for keys: {keys}")
            return result
        except Exception as e:
            log_context.error(f"Redis DELETE failed for keys {keys}: {e}")
            return 0

    def exists(self, *keys: str) -> int:
        """Check if one or more keys exist in Redis.

        Args:
            *keys (str): One or more Redis keys to check

        Returns:
            int: Number of keys that exist
        """
        try:
            result = self.client.exists(*keys)
            log_context.debug(f"Redis EXISTS check for keys: {keys}")
            return result
        except Exception as e:
            log_context.error(f"Redis EXISTS failed for keys {keys}: {e}")
            return 0

    def expire(self, key: str, ttl: int) -> bool:
        """Set TTL for an existing key.

        Args:
            key (str): The Redis key
            ttl (int): Time to live in seconds

        Returns:
            bool: True if the operation was successful, False otherwise
        """
        try:
            result = self.client.expire(key, ttl)
            log_context.debug(f"Redis EXPIRE successful for key: {key}")
            return bool(result)
        except Exception as e:
            log_context.error(f"Redis EXPIRE failed for key {key}: {e}")
            return False

    def ttl(self, key: str) -> int:
        """Get the TTL of a key.

        Args:
            key (str): The Redis key

        Returns:
            int: TTL in seconds, -1 if no TTL, -2 if key doesn't exist
        """
        try:
            result = self.client.ttl(key)
            log_context.debug(f"Redis TTL check for key: {key}")
            return result
        except Exception as e:
            log_context.error(f"Redis TTL failed for key {key}: {e}")
            return -2

    def flush_db(self) -> bool:
        """Flush the current Redis database.

        Returns:
            bool: True if the operation was successful, False otherwise
        """
        try:
            result = self.client.flushdb()
            log_context.warning("Redis database flushed")
            return bool(result)
        except Exception as e:
            log_context.error(f"Redis FLUSHDB failed: {e}")
            return False

    def keys(self, pattern: str = "*") -> list[str]:
        """Get all keys matching a pattern.

        Args:
            pattern (str): The pattern to match (default: "*" for all keys)

        Returns:
            List[str]: List of matching keys
        """
        try:
            keys = self.client.keys(pattern)
            # Decode bytes keys to strings
            result = [
                key.decode("utf-8") if isinstance(key, bytes) else key for key in keys
            ]
            log_context.debug(
                f"Redis KEYS found {len(result)} keys matching pattern: {pattern}"
            )
            return result
        except Exception as e:
            log_context.error(f"Redis KEYS failed for pattern {pattern}: {e}")
            return []

    def close(self) -> None:
        """Close all connections in the pool.

        This method closes the connection pool and cleans up resources.
        """
        try:
            self.connection_pool.disconnect()
            log_context.info("Redis connection pool closed")
        except Exception as e:
            log_context.error(f"Error closing Redis connection pool: {e}")


# Filter out None values from Redis config
filtered_redis_config = {k: v for k, v in REDIS_CONFIG.items() if v is not None}

# Create a global instance of the Redis connection pool
redis_pool = RedisPool(POOL_SIZE, **filtered_redis_config)



from benchmark.tau_bench.envs.base import Env
from benchmark.tau_bench.envs.user import UserStrategy


def get_env(
    env_name: str,
    user_strategy: str | UserStrategy,
    user_model: str,
    task_split: str,
    user_provider: str | None = None,
    task_index: int | None = None,
) -> Env:
    if env_name == "retail":
        from benchmark.tau_bench.envs.retail import MockRetailDomainEnv

        return MockRetailDomainEnv(
            user_strategy=user_strategy,
            user_model=user_model,
            task_split=task_split,
            user_provider=user_provider,
            task_index=task_index,
        )
    elif env_name == "airline":
        from benchmark.tau_bench.envs.airline import MockAirlineDomainEnv

        return MockAirlineDomainEnv(
            user_strategy=user_strategy,
            user_model=user_model,
            task_split=task_split,
            user_provider=user_provider,
            task_index=task_index,
        )
    else:
        raise ValueError(f"Unknown environment: {env_name}")



from .calculate import Calculate
from .cancel_pending_order import CancelPendingOrder
from .exchange_delivered_order_items import ExchangeDeliveredOrderItems
from .find_user_id_by_email import FindUserIdByEmail
from .find_user_id_by_name_zip import FindUserIdByNameZip
from .get_order_details import GetOrderDetails
from .get_product_details import GetProductDetails
from .get_user_details import GetUserDetails
from .list_all_product_types import ListAllProductTypes
from .modify_pending_order_address import ModifyPendingOrderAddress
from .modify_pending_order_items import ModifyPendingOrderItems
from .modify_pending_order_payment import ModifyPendingOrderPayment
from .modify_user_address import ModifyUserAddress
from .return_delivered_order_items import ReturnDeliveredOrderItems
from .think import Think
from .transfer_to_human_agents import TransferToHumanAgents

ALL_TOOLS = [
    Calculate,
    CancelPendingOrder,
    ExchangeDeliveredOrderItems,
    FindUserIdByEmail,
    FindUserIdByNameZip,
    GetOrderDetails,
    GetProductDetails,
    GetUserDetails,
    ListAllProductTypes,
    ModifyPendingOrderAddress,
    ModifyPendingOrderItems,
    ModifyPendingOrderPayment,
    ModifyUserAddress,
    ReturnDeliveredOrderItems,
    Think,
    TransferToHumanAgents,
]



# Go to the parent folder of this file (calendar), then Run python -m unittest test.py to test the code in this file.
import json
import os
import sys
import unittest
from typing import Any

from arklex.env.env import Env
from arklex.orchestrator.orchestrator import AgentOrg
from arklex.utils.logging_utils import LogContext

# May not need after pip install agentorg
sys.path.insert(0, os.path.abspath("../../"))
print(sys.path)

log_context = LogContext(__name__)


class Logic_Test(unittest.TestCase):
    file_path: str = "test_cases.json"
    with open(file_path, encoding="UTF-8") as f:
        TEST_CASES: list[dict[str, Any]] = json.load(f)

    @classmethod
    def setUpClass(cls) -> None:
        """Method to prepare the test fixture. Run BEFORE the test methods."""
        cls.user_prefix: str = "user"
        cls.worker_prefix: str = "assistant"
        file_path: str = "taskgraph.json"
        with open(file_path, encoding="UTF-8") as f:
            cls.config: dict[str, Any] = json.load(f)
        cls.env: Env = Env(
            tools=cls.config.get("tools", []),
            workers=cls.config.get("workers", []),
            slotsfillapi=cls.config["slotfillapi"],
        )
        cls.config["input_dir"] = "./"
        os.environ["DATA_DIR"] = cls.config["input_dir"]

    @classmethod
    def tearDownClass(cls) -> None:
        """Method to tear down the test fixture. Run AFTER the test methods."""

    def _get_api_bot_response(
        self, user_text: str, history: list[dict[str, str]], params: dict[str, Any]
    ) -> tuple[str, dict[str, Any]]:
        data: dict[str, Any] = {
            "text": user_text,
            "chat_history": history,
            "parameters": params,
        }
        orchestrator = AgentOrg(config=self.config, env=self.env)
        result: dict[str, Any] = orchestrator.get_response(data)

        return result["answer"], result["parameters"]

    def test_Unittest0(self) -> None:
        log_context.info("\n=============Unit Test 0=============")
        log_context.info(f"{self.TEST_CASES[0]['description']}")
        history: list[dict[str, str]] = []
        params: dict[str, Any] = {}
        nodes: list[str] = []
        for node in self.config["nodes"]:
            if node[1].get("type", "") == "start":
                start_message: str = node[1]["attribute"]["value"]
                break
        history.append({"role": self.worker_prefix, "content": start_message})

        for user_text in self.TEST_CASES[0]["user_utterance"]:
            print(f"User: {user_text}")
            output, params = self._get_api_bot_response(user_text, history, params)
            print(f"Bot: {output}")
            nodes.append(params.get("curr_node"))
            history.append({"role": self.user_prefix, "content": user_text})
            history.append({"role": self.worker_prefix, "content": output})
        print(f"{self.TEST_CASES[0]['criteria']}")
        # print(json.dumps(history, indent=4))
        self.assertEqual(nodes, self.TEST_CASES[0]["trajectory"])

    def test_Unittest1(self) -> None:
        print("\n=============Unit Test 1=============")
        print(f"{self.TEST_CASES[1]['description']}")
        history: list[dict[str, str]] = []
        params: dict[str, Any] = {}
        nodes: list[str] = []
        for node in self.config["nodes"]:
            if node[1].get("type", "") == "start":
                start_message: str = node[1]["attribute"]["value"]
                break
        history.append({"role": self.worker_prefix, "content": start_message})

        for user_text in self.TEST_CASES[1]["user_utterance"]:
            print(f"User: {user_text}")
            output, params = self._get_api_bot_response(user_text, history, params)
            print(f"Bot: {output}")
            nodes.append(params.get("curr_node"))
            history.append({"role": self.user_prefix, "content": user_text})
            history.append({"role": self.worker_prefix, "content": output})
        print(f"{self.TEST_CASES[1]['criteria']}")
        # print(json.dumps(history, indent=4))
        self.assertEqual(nodes, self.TEST_CASES[1]["trajectory"])

    def test_Unittest2(self) -> None:
        print("\n=============Unit Test 2=============")
        print(f"{self.TEST_CASES[2]['description']}")
        history: list[dict[str, str]] = []
        params: dict[str, Any] = {}
        nodes: list[str] = []
        for node in self.config["nodes"]:
            if node[1].get("type", "") == "start":
                start_message: str = node[1]["attribute"]["value"]
                break
        history.append({"role": self.worker_prefix, "content": start_message})

        for user_text in self.TEST_CASES[2]["user_utterance"]:
            print(f"User: {user_text}")
            output, params = self._get_api_bot_response(user_text, history, params)
            print(f"Bot: {output}")
            nodes.append(params.get("curr_node"))
            history.append({"role": self.user_prefix, "content": user_text})
            history.append({"role": self.worker_prefix, "content": output})
        print(f"{self.TEST_CASES[2]['criteria']}")
        # print(json.dumps(history, indent=4))
        self.assertEqual(nodes, self.TEST_CASES[2]["trajectory"])


if __name__ == "__main__":
    unittest.main()



"""Tool test utilities for the Arklex framework.

This module provides test utilities for validating tool behavior in the Arklex
framework. It includes test orchestrators for Shopify tools, with validation
methods to ensure correct task graph paths and node status.
"""

import json
import os
from typing import Any

from arklex.orchestrator.NLU.core.slot import SlotFiller
from arklex.orchestrator.NLU.services.model_service import DummyModelService
from arklex.utils.logging_utils import LogContext
from tests.utils.utils import MockOrchestrator, MockResourceInitializer

log_context = LogContext(__name__)


class ShopifyToolOrchestrator(MockOrchestrator):
    def __init__(self, config_file_path: str) -> None:
        """Initialize the Shopify tool orchestrator.

        Args:
            config_file_path (str): Path to the configuration file.
        """
        fixed_args: str = os.environ.get("SHOPIFY_FIXED_ARGS", "{}")
        # Handle case where environment variable might be empty or None
        if not fixed_args or fixed_args.strip() == "":
            fixed_args = "{}"
        try:
            self.fixed_args: dict[str, Any] = json.loads(fixed_args)
        except json.JSONDecodeError:
            # If JSON parsing fails, use empty dict as fallback
            self.fixed_args = {}
        super().__init__(config_file_path, self.fixed_args)
        self.resource_initializer = MockResourceInitializer()

    def _validate_result(
        self,
        test_case: dict[str, Any],
        history: list[dict[str, str]],
        params: dict[str, Any],
    ) -> None:
        """Validate the test results for Shopify tools.

        This function only checks that the assistant's response is non-empty.
        """
        assistant_records: list[dict[str, str]] = [
            message for message in history if message["role"] == "assistant"
        ]
        for record in assistant_records:
            assert record["content"] != ""

    def initialize_slotfillapi(self, slotsfillapi: str) -> SlotFiller:
        """Initialize the slot filling API.

        Args:
            slotsfillapi: API endpoint for slot filling

        Returns:
            Initialized SlotFiller instance
        """
        dummy_config = {
            "model_name": "dummy",
            "api_key": "dummy",
            "endpoint": "http://dummy",
            "model_type_or_path": "dummy-path",
            "llm_provider": "dummy",
        }
        if not isinstance(slotsfillapi, str):
            log_context.error("slotsfillapi must be a string")
            return None
        if not slotsfillapi:
            log_context.warning(
                "slotsfillapi is empty, using local model-based slot filling"
            )
            return SlotFiller(DummyModelService(dummy_config))
        log_context.info(f"Initializing SlotFiller with API URL: {slotsfillapi}")
        return SlotFiller(slotsfillapi)
